#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List, NoReturn

import numpy as np
import networkx as nx

from joblib import delayed, Parallel

from ase import Atoms

from ..builder.group import create_an_intersect_group
from ..graph.creator import StruGraphCreator
from ..graph.comparison import paragroup_unique_chem_envs
from ..utils.command import CustomTimer

from .selector import AbstractSelector


def single_create_structure_graph(
    atoms: Atoms, graph_params: dict, group_commands: List[str] = None
) -> List[nx.Graph]:
    """Create structure graph and get selected chemical environments.

    Find atoms with selected chemical symbols or in the defined region.

    Args:
        graph_params: Parameters for the graph representation.
        spec_params: Selected criteria.
        atoms: Input structure.

    Returns:
        A list of graphs that represent the chemical environments of selected atoms.

    """
    stru_creator = StruGraphCreator(**graph_params)

    # - find atoms whose environments need to extract
    ads_indices = create_an_intersect_group(atoms, group_commands)

    _ = stru_creator.generate_graph(atoms, ads_indices_=ads_indices)

    chem_envs = stru_creator.extract_chem_envs(atoms)

    return chem_envs


class GraphSelector(AbstractSelector):

    name = "graph"

    default_parameters = dict(
        group_commands=None,
        pbc_grid=[2, 2, 0],
        graph_radius=2,
        neigh_params=dict(covalent_ratio=1.1, skin=0.0),
    )

    def __init__(self, directory="./", *args, **kwargs):
        """"""
        super().__init__(directory=directory, *args, **kwargs)

        # -- check params
        if self.group_commands is None:
            raise RuntimeError(f"{self.name} selector need group_commands.")

        return

    def _select_indices(self, frames: List[Atoms], *args, **kwargs) -> List[int]:
        """"""
        nframes = len(frames)

        selected_indices = []

        graph_params = copy.deepcopy(
            dict(
                adsorbate_elements=[],
                pbc_grid=self.pbc_grid,
                graph_radius=self.graph_radius,
                neigh_params=self.neigh_params,
            )
        )
        group_commands = copy.deepcopy(self.group_commands)

        # - create graphs
        with CustomTimer(name="create-graphs", func=self._print):
            ret = Parallel(n_jobs=self.njobs)(
                delayed(single_create_structure_graph)(a, graph_params, group_commands)
                for a in frames
            )

        # - check if the ret is empty
        ret_envs = []
        for x in ret:
            ret_envs.extend(x)

        if ret_envs:
            self._print("Typical Chemical Environment " + str(ret_envs[0]))
            ret_env_groups = ret
            with CustomTimer(name="check-uniqueness", func=self._print):
                unique_envs, unique_groups = paragroup_unique_chem_envs(
                    ret_env_groups,
                    list(enumerate(frames)),
                    directory=self.directory,
                    n_jobs=1,
                )
            selected_indices = [x[0][0] for x in unique_groups]
        else:
            self._print("Cant find valid atoms for comparison...")
            selected_indices = list(range(nframes))

        self._write_results(frames, selected_indices, unique_groups=unique_groups)

        return selected_indices

    def _write_results(self, frames, selected_indices, *args, **kwargs) -> NoReturn:
        """Write selection results into file that can be used for restart."""
        data = []
        for s in selected_indices:
            atoms = frames[s]
            # - gather info
            confid = atoms.info.get("confid", -1)
            natoms = len(atoms)
            ae = atoms.get_potential_energy() / natoms
            maxforce = np.max(np.fabs(atoms.get_forces(apply_constraint=True)))
            data.append([s, confid, natoms, ae, maxforce])

        # NOTE: output index and confid maybe inconsistent since the frames are
        #       sorted in advance...
        if data:
            np.savetxt(
                self.info_fpath,
                data,
                fmt="%8d  %8d  %8d  %12.4f  %12.4f",
                # fmt="{:>8d}  {:>8d}  {:>8d}  {:>12.4f}  {:>12.4f}",
                header="{:>6s}  {:>8s}  {:>8s}  {:>12s}  {:>12s}".format(
                    *"index confid natoms AtomicEnergy MaxForce".split()
                ),
            )
        else:
            np.savetxt(
                self.info_fpath,
                [[np.NaN]] * 5,
                header="{:>6s}  {:>8s}  {:>8s}  {:>12s}  {:>12s}".format(
                    *"index confid natoms AtomicEnergy MaxForce".split()
                ),
            )

        # -- unique info
        unique_groups = kwargs.get("unique_groups", None)
        if unique_groups is not None:
            unique_data = []
            for i, x in enumerate(unique_groups):
                data = ["ug" + str(i)]
                data.extend([a[0] for a in x])
                unique_data.append(data)
            content = "# unique, indices\n"
            content += "# ncandidates {}\n".format(len(selected_indices))
            for d in unique_data:
                content += ("{:<8s}  " + "{:<8d}  " * (len(d) - 1) + "\n").format(*d)

            unique_info_path = self.info_fpath.parent / (
                self.info_fpath.stem + "-extra.txt"
            )
            with open(unique_info_path, "w") as fopen:
                fopen.write(content)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib

import numpy as np

from ase.units import kB
from ase.io import read, write


def compute_diff(endiffs, temp=600):
    """"""
    beta = 1./(kB*temp)
    print(beta)
    print(np.average(np.exp(-beta*endiffs), axis=1))

    return (-beta**-1)*np.log(np.average(np.exp(-beta*endiffs), axis=1))


def reweight():
    frames = read(f"./post/bin_{step}.xyz", ":")
        
    nframes = len(frames)
    print(f"nframes: {nframes}")
    energies_dft = np.array([a.get_potential_energy() for a in frames])
    
    dp_frames = read(f"./post/bin_{step}/_data/2194d49907edfec2363c7f2619662c62_cache.xyz", ":")
    #dp_frames = read("./post/bin_5/_data/f62ddad7513a99cffa4fac6a3712fe46_cache.xyz", ":")
    nframes_dp = len(dp_frames)
    energies_dp = np.array([a.get_potential_energy() for a in dp_frames])
    print(energies_dft-energies_dp)
    
    assert nframes == nframes_dp
    
    #endiff = energies_dft - energies_dp
    endiff = np.array([0.3, 0.2, 0.1, 0.0, -0.1, -0.2, -0.3])
    
    print(compute_diff(endiff))

    return


wdir = pathlib.Path("./_spc/")
post = pathlib.Path("./post/")

steps = range(20)
binwidth = 64

for step in steps:
    print(f"step: {step}")
    if not (post/"bins"/f"bin_{step}.xyz").exists():
        confids = [i+step*binwidth for i in range(binwidth)]
        if any([not (wdir/f"cand{i}").exists() for i in confids]):
            print("NOT FINISHED!!!")
            continue
        
        frames = []
        for i in confids:
            frames.append(read(wdir/f"cand{i}"/"vasprun.xml"))
        write(post/"bins"/f"bin_{step}.xyz", frames)
    else:
        print("EXTRACTED !!!")

print("=====")

endiffs = []
for step in steps:
    print(f"step: {step}")
    if (post/f"bin_{step}"/"results"/"end_frames.xyz").exists():
        #print("FOUND !!!")
        dft_frames = read(post/"bins"/f"bin_{step}.xyz", ":")
        dft_energies = np.array([a.get_potential_energy() for a in dft_frames])
        dp_frames = read(post/f"bin_{step}"/"results"/"end_frames.xyz", ":")
        dp_energies = np.array([a.get_potential_energy() for a in dp_frames])
        endiffs.append(dft_energies - dp_energies)
    else:
        print("ZEROD!!!")
        endiffs.append(np.zeros((binwidth,)))
endiffs = np.array(endiffs)
np.savetxt("./diff.txt", endiffs, fmt="%8.4f")
#print(endiffs)
print("avg: ", np.average(endiffs, axis=1))
print("std: ", np.sqrt(np.var(endiffs, axis=1)))

weights = compute_diff(endiffs, temp=600)
for i in [4, 9, 13, 14, 19]:
    weights[i] = np.nan
print(weights)


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import pathlib
from typing import Union, List

import numpy as np

import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

import ase
from ase import Atoms
from ase.io import read, write

from .validator import AbstractValidator
from .utils import wrap_traj, smooth_curve
from ..data.array import AtomsNDArray

"""This module is used to analyse the particle distribution in the system.
"""


def plot_mass_distribution(wdir, prefix, bincentres, mass, title):
    """"""

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(16,12))

    plt.suptitle("Mass Distribution Function")

    ax.set_xlabel("R [Å]")
    ax.set_ylabel("Mass [a.u./Å^2]")
    ax.set_title(title)

    avg_mass = np.average(mass, axis=0)
    bincentres_, avg_mass_ = smooth_curve(bincentres, avg_mass)
    ax.plot(bincentres_, avg_mass_, label="avg")

    std_mass = np.sqrt(np.var(mass, axis=0))
    bincentres_, lo_mass_ = smooth_curve(bincentres, avg_mass-std_mass)
    bincentres_, hi_mass_ = smooth_curve(bincentres, avg_mass+std_mass)
    ax.fill_between(bincentres_, lo_mass_, hi_mass_, alpha=0.2, label="${\sigma}$")

    # - stepwise
    #nframes = rdf.shape[0]
    #for i in range(0, nframes, step):
    #    cur_rdf = rdf[i]
    #    ax.plot(bincentres, cur_rdf, label=f"rdf{i}")

    ax.legend()

    plt.savefig(wdir/f"{prefix}mdf.png")

    data = np.vstack((bincentres_, avg_mass_, lo_mass_, hi_mass_)).T

    np.savetxt(
        wdir/f"{prefix}mdf.dat", data, 
        fmt=("%12.4f  "*4),
        header=("{:<8s}  "*4).format("r", "avg", "avg-sigma", "avg+sigma")
    )

    return


class MassDistributionValidator(AbstractValidator):

    """Validate mass distribution.

    TODO:
        Support molecular mass distribution by tags.

    """

    def __init__(self, symbols: List[str], drange, vector=[0,0,1], nbins=60, directory: Union[str, pathlib.Path] = "./", *args, **kwargs):
        """Init a mass distribution validator.

        Args:
            symbols:

        """
        super().__init__(directory, *args, **kwargs)

        self.symbols = symbols
        self.vector = np.array(vector) / np.linalg.norm(vector)
        self.nbins = nbins

        self.rmin, self.rmax = drange

        return
    
    def _process_data(self, data) -> List[List[Atoms]]:
        """"""
        data = AtomsNDArray(data)

        if data.ndim == 1:
            data = [data.tolist()]
        elif data.ndim == 2: # assume it is from extract_cache...
            data = data.tolist()
        elif data.ndim == 3: # assume it is from a compute node...
            data_ = []
            for d in data[:]: # TODO: add squeeze method?
                data_.extend(d)
            data = data_
        else:
            raise RuntimeError(f"Invalid shape {data.shape}.")

        return data
    
    def run(self, dataset, worker=None, *args, **kwargs):
        """Process reference and prediction data separately.

        TODO:

            Support average over several trajectories.

        """
        self._print("process reference ->")
        reference = dataset.get("reference", None)
        if reference is not None:
            self._irun(self._process_data(reference)[0], prefix="ref-")

        self._print("process prediction ->")
        prediction = dataset.get("prediction", None)
        if prediction is not None:
            self._irun(self._process_data(prediction)[0], prefix="pre-")

        return

    def _irun(self, frames: List[Atoms], prefix):
        """"""
        #pmin = np.floor(np.min(projected_distances))
        #pmax = np.ceil(np.max(projected_distances))
        bin_edges = np.linspace(self.rmin, self.rmax, self.nbins, endpoint=False).tolist()
        bin_edges.append(self.rmax)
        bin_edges = np.array(bin_edges)
        #print(len(bin_edges), bin_edges)
        bincentres = (bin_edges[:-1]+bin_edges[1:])/2.
        #print("bincentres: ", bincentres)

        # NOTE: only support structure with fixed cell
        atoms = frames[0]
        cell = atoms.get_cell(complete=True)
        surf_area = np.dot(np.cross(cell[0], cell[1]), self.vector)

        mass = []
        for atoms in frames:
            curr_mass = self._compute_mass_distribution(
                atoms, bin_edges, self.nbins
            )
            mass.append(curr_mass)
        #avg_mass = np.mean(mass, axis=0)
        #min_mass = np.min(mass, axis=0)
        #max_mass = np.max(mass, axis=0)

        mass = np.array(mass) / surf_area # normalised by surface area

        plot_mass_distribution(self.directory, prefix, bincentres, mass, ", ".join(self.symbols))

        return
    
    def _compute_mass_distribution(self, atoms: List[Atoms], bin_edges, nbins: int):
        """"""
        selected_indices = []
        for i, a in enumerate(atoms):
            if a.symbol in self.symbols:
                selected_indices.append(i)
        selected_positions = copy.deepcopy(atoms.get_positions()[selected_indices, :])
        projected_distances = [np.dot(pos, self.vector) for pos in selected_positions]

        masses = atoms.get_masses()
        selected_masses = [masses[i] for i in selected_indices]

        # - create bins
        bin_indices = np.digitize(projected_distances, bin_edges, right=False)

        groups = [[] for i in range(nbins)] # mass
        for i, i_bin in enumerate(bin_indices):
            # dump prop not in pmin and pmax
            if i_bin <= nbins:
                groups[i_bin-1].append(selected_masses[i])
        mass_by_digit = np.array([np.sum(x) for x in groups])

        #print("groups: ", groups)
        #print("mas: ", mass_by_digit)

        return mass_by_digit


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

import subprocess

from .scheduler import AbstractScheduler


class LsfScheduler(AbstractScheduler):
    """Load Sharing Facility (LSF) scheduler.

    A LSF scheduler. Commands are bjobs, bsub, bkill.

    """

    name = "lsf"

    PREFIX = "#BSUB"
    SUFFIX = ".lsf"
    SHELL = "#!/bin/bash -l"

    SUBMIT_COMMAND = "bsub < "
    # ENQUIRE_COMMAND = "`which squeue` -u `whoami` --format=\"%.12i %.12P %.60j %.4t %.12M %.12L %.5D %.4C\""
    ENQUIRE_COMMAND = "`which bjobs` -u `whoami` -w"

    # compability for different machine
    # not all keywords are necessary
    default_parameters = {
        "J": "lsfJob",  # job name
        "q": None,  # partition
        "W": None,  # time
        # - CPU
        "R": None,  # nodes
        "n": None,  # ncpus
        "M": None,  # memory
        "o": "lsf.o%J",  # output
        "e": "lsf.e%J",  # error
        # - GPU
        # "gres": None,
        # "mem-per-gpu": None, # "32G"
    }

    running_status = ["RUN", "PEND", "DONE", "EXIT", "SSUSP", "USUSP"]

    def __str__(self) -> str:
        """Return the content of the job script."""
        # - scheduler params
        content = self.SHELL + "\n"
        for key, value in self.parameters.items():
            if value:
                # content += "{} --{}={}\n".format(self.PREFIX, key, value)
                content += "{} -{} {}\n".format(self.PREFIX, key, value)
            # else:
            #    raise ValueError("Keyword *%s* not properly set." %key)

        if self.environs:
            content += "\n\n"
            content += self.environs

        if self.user_commands:
            content += "\n\n"
            content += self.user_commands

        return content

    @AbstractScheduler.job_name.setter
    def job_name(self, job_name_: str):
        self._job_name = job_name_
        self.set(**{"J": self._job_name})
        return

    def is_finished(self) -> bool:
        """Check if the job were finished.

        Currently, we only check whether the job name has been found in the queue.
        If found, we assume the job is still running. Otherwise, it may have already
        finished.

        Returns:
            Whether the jobs is finished. True for finished, False otherwise.

        """
        # - run enquire
        p = subprocess.Popen(
            [self.ENQUIRE_COMMAND],
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            universal_newlines=True,
        )
        fout = p.stdout
        lines = fout.readlines()

        # - run over results
        """format
        JOBID   USER    STAT  QUEUE      FROM_HOST   EXEC_HOST   JOB_NAME   SUBMIT_TIME
        120727  jxu     PEND  normal     manage01                *ee.script Mar  1 21:55
        """
        finished = False
        for line in lines[1:]:  # skipe first info line
            data = line.strip().split()
            jobid, name, status = data[0], data[6], data[2]
            # if name.startswith(self.prefix) and status in self.running_status:
            #    indices = re.match(self.prefix+"*", name).span()
            #    if indices is not None:
            #        confid = int(name[indices[1]:])
            #    confids.append(int(confid))
            if name == self.job_name:
                finished = False
                break
        else:
            finished = True

        return finished


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List, Tuple


import numpy as np

from ase import Atoms


def get_distance_and_shift(cell, positions, i, j, pbc: bool=True):
    """"""
    # FIXME: ...

    return


def compute_distance_and_shift(atoms: Atoms, neighlist, grp_a: List[int], grp_b: List[int]):
    """Find valid bond pairs between group_a and group_b.

    NOTE: We use a full-list here.

    """
    # -
    cell = atoms.get_cell(complete=True)

    # - find pairs within given distance
    bond_pairs = []
    bond_distances = []
    bond_shifts = []
    for i in grp_a:
        indices, offsets = neighlist.get_neighbors(i)
        for j, offset in zip(indices, offsets):
            if j in grp_b:
                shift = np.dot(offset, cell)
                dis = np.linalg.norm(atoms.positions[i] - (atoms.positions[j] + shift))
                bond_pairs.append(sorted([i, j]))
                bond_distances.append(dis)
                bond_shifts.append(dis)

    bond_pairs = np.array(bond_pairs, dtype=np.int32)
    bond_distances = np.array(bond_distances)
    # bond_shifts = np.array(bond_shifts)

    return bond_pairs, bond_distances, bond_shifts


def get_bond_information(
    atoms: Atoms,
    neighlist,
    eqdis_dict: dict,
    covalent_min: float,
    target_indices: List[int],
    allowed_bonds: List[Tuple[str, str]],
):
    """Find valid bond pairs and compute their strains.

    NOTE: We use a half-list here. This is used in bond-boost.

    """
    # -
    cell = atoms.get_cell(complete=True)

    # - find pairs within given distance
    bond_pairs = []
    bond_curr_distances = []
    bond_curr_shifts = []
    bond_equi_distances = []
    for i in target_indices:
        sym_i = atoms[i].symbol
        indices, offsets = neighlist.get_neighbors(i)
        for j, offset in zip(indices, offsets):
            sym_j = atoms[j].symbol
            pair = (sym_i, sym_j)
            if pair in allowed_bonds:
                shift = np.dot(offset, cell)
                dis = np.linalg.norm(
                    atoms.positions[i] - (atoms.positions[j] + shift)
                )
                if dis >= eqdis_dict[pair] * covalent_min:
                    bond_pairs.append(tuple(sorted([i, j])))
                    bond_curr_distances.append(dis)
                    bond_curr_shifts.append(shift)
                    bond_equi_distances.append(eqdis_dict[pair])

    bond_pairs = bond_pairs
    bond_curr_distances = np.array(bond_curr_distances)
    bond_equi_distances = np.array(bond_equi_distances)

    return bond_pairs, bond_curr_distances, bond_curr_shifts, bond_equi_distances


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import subprocess

from .scheduler import AbstractScheduler


class SlurmScheduler(AbstractScheduler):
    """Slurm scheduler.

    A slurm scheduler.

    """

    name = "slurm"

    PREFIX = "#SBATCH"
    SUFFIX = ".slurm"
    SHELL = "#!/bin/bash -l"

    SUBMIT_COMMAND = "sbatch"
    ENQUIRE_COMMAND = '`which squeue` -u `whoami` --format="%.12i %.12P %.60j %.4t %.12M %.12L %.5D %.4C"'

    # compability for different machine
    # not all keywords are necessary
    default_parameters = {
        "job-name": "slurmJob",
        "account": None,
        "partition": None,
        "time": None,
        # - CPU
        "nodes": None,
        "ntasks": None,
        "tasks-per-node": None,
        "cpus-per-task": None,
        "mem-per-cpu": None,  # "4G"
        # "output": "slurm.o%j",
        # "error": "slurm.e%j"
        # - GPU
        "gres": None,
        "mem-per-gpu": None,  # "32G"
    }

    running_status = ["R", "Q", "PD", "CG"]

    def __str__(self) -> str:
        """Return the content of the job script."""
        # - slurm params
        content = self.SHELL + "\n"
        for key, value in self.parameters.items():
            if value:
                content += "{} --{}={}\n".format(self.PREFIX, key, value)
            # else:
            #    raise ValueError("Keyword *%s* not properly set." %key)

        if self.environs:
            content += "\n\n"
            content += self.environs

        if self.user_commands:
            content += "\n\n"
            content += self.user_commands

        return content

    @AbstractScheduler.job_name.setter
    def job_name(self, job_name_: str):
        self._job_name = job_name_
        self.set(**{"job-name": self._job_name})
        return

    def is_finished(self) -> bool:
        """Check if the job were finished.

        Currently, we only check whether the job name has been found in the queue.
        If found, we assume the job is still running. Otherwise, it may have already
        finished.

        Returns:
            Whether the jobs is finished. True for finished, False otherwise.

        """
        # - run enquire
        p = subprocess.Popen(
            [self.ENQUIRE_COMMAND],
            shell=True,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            close_fds=True,
            universal_newlines=True,
        )
        fout = p.stdout
        lines = fout.readlines()

        # - run over results
        finished = False
        for line in lines[1:]:  # skipe first info line
            data = line.strip().split()
            jobid, name, status = data[0], data[2], data[3]
            # if name.startswith(self.prefix) and status in self.running_status:
            #    indices = re.match(self.prefix+"*", name).span()
            #    if indices is not None:
            #        confid = int(name[indices[1]:])
            #    confids.append(int(confid))
            if name == self.parameters["job-name"]:
                finished = False
                break
        else:
            finished = True

        return finished


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np

import jax
import jax.numpy as jnp


class DistanceColvar():

    def __init__(self, pairs, *args, **kwargs) -> None:
        """"""
        self.params = np.array(pairs).T

        return

    @staticmethod
    @jax.jit
    def cvfunc(positions, params):
        """"""
        pair_indices = params

        pair_positions = jnp.take(positions, pair_indices, axis=0)
        dvecs = pair_positions[0] - pair_positions[1]
        distances = jnp.linalg.norm(dvecs, axis=1)

        return distances[jnp.newaxis, :] # (1, num_dim)


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import Optional

import numpy as np

from ase import Atoms
from ase import units
from ase.constraints import FixAtoms
from ase.md.md import MolecularDynamics
from ase.md.verlet import VelocityVerlet
from ase.md.langevin import Langevin


class TimeStampedMonteCarlo(MolecularDynamics):

    def __init__(
        self,
        atoms: Atoms,
        maxstepsize: float = 0.20,
        fixcm: bool = True,  # TODO:
        temperature: Optional[float] = None,
        temperature_K: Optional[float] = None,
        rng=np.random.Generator(np.random.PCG64()),
        *args,
        **kwargs,
    ):
        """Time-stamped force-bias Monte Carlo.

        Args:
            atoms: Atoms object.

        """
        MolecularDynamics.__init__(self, atoms=atoms, *args, **kwargs)

        self.maxstepsize = maxstepsize

        self.fix_com = fixcm

        self.temp = units.kB * self._process_temperature(
            temperature, temperature_K, "eV"
        )  # == kbT

        self.rng = rng

        # - get minimum mass
        #   no more new atoms should be added after the simulation is defined.
        self.mass_min = np.min(self.masses)

        return

    def step(self, forces=None):
        """"""
        atoms = self.atoms

        # NOTE: The md forces does not apply FixAtoms, 
        #       so all forces are physical
        if forces is None:
            # forces = atoms.get_forces(apply_constraint=True, md=False)
            forces = atoms.get_forces(apply_constraint=True, md=True)
        # print(f"forces: {forces}")

        # - get displacement steps
        stepsizes = self.maxstepsize * np.power(self.mass_min / self.masses, 0.25)

        natoms = len(atoms)
        displacements = np.zeros((natoms, 3))

        for i in range(natoms):
            for j in range(3):
                P_acc, P_ran = 0, 1
                gamma = forces[i, j] * stepsizes[i] / (2.0 * self.temp)
                gamma_exp = np.exp(gamma)
                gamma_expi = 1.0 / gamma_exp
                while P_acc < P_ran:
                    xi = 2.0 * self.rng.uniform(-1.0, 1.0)
                    P_ran = self.rng.uniform()
                    if xi < 0:
                        P_acc = np.exp(2.0 * xi * gamma) * gamma_exp - gamma_expi
                        P_acc = P_acc / (gamma_exp - gamma_expi)
                    elif xi > 0:
                        P_acc = gamma_expi - np.exp(2.0 * xi * gamma) * gamma_expi
                        P_acc = P_acc / (gamma_exp - gamma_expi)
                    else:
                        P_acc = 1.0

                    displacements[i][j] = xi * stepsizes[i]

        positions = atoms.get_positions()

        # print(f"prev_com: {atoms.get_center_of_mass()}")

        # - remove translation
        if self.fix_com:
            for constraint in atoms.constraints:
                if isinstance(constraint, FixAtoms):
                    move_indices = [
                        i for i in range(natoms) if i not in constraint.index
                    ]
                    break
            else:
                move_indices = list(range(natoms))
            com_disp = (
                self.masses[move_indices].T
                @ displacements[move_indices, :]
                / np.sum(self.masses[move_indices])
            )
            displacements[move_indices, :] -= com_disp
            # prev_com = atoms.get_center_of_mass()

        # - remove rotation
        ...

        # - adjust posistions
        atoms.set_positions(positions + displacements)

        # print(f"move_com: {atoms.get_center_of_mass()}")

        # NOTE: ase v3.22.1 has a bug in set_center_of_mass
        #       so we do it by ourselves
        # NOTE: set_com should consider constraints!!!
        # if self.fix_com:
        #     diff_com = prev_com - atoms.get_center_of_mass()
        #     print(f"diff_com: {diff_com}")
        #     print(f"move_com: {atoms.get_center_of_mass()}")
        #     atoms.set_positions(
        #         atoms.get_positions() + diff_com, apply_constraint=False
        #     )

        # print(f"curr_com: {atoms.get_center_of_mass()}")

        return forces


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import copy
from pathlib import Path
from typing import List

import yaml

from ase.calculators.calculator import Calculator

from .. import AbstractPotentialManager, AbstractTrainer


class BeannTrainer(AbstractTrainer):
    
    name = "beann"
    command = ""
    freeze_command = ""
    prefix = "config"

    def __init__(
        self, config: dict, type_list: List[str] = None, train_epochs: int = 200, 
        directory=".", command="train", freeze_command="freeze", 
        random_seed: int = 1112, *args, **kwargs
    ) -> None:
        super().__init__(
            config=config, type_list=type_list, train_epochs=train_epochs, 
            directory=directory, command=command, freeze_command=freeze_command, 
            random_seed=random_seed, *args, **kwargs
        )

        return
    
    def _resolve_train_command(self, *args, **kwargs):
        """python -u /users/jxu/repository/EANN/eann --config ./config.yaml train"""

        return
    
    def _resolve_freeze_command(self, *args, **kwargs):
        """python -u /users/jxu/repository/EANN/eann --config ./config.yaml freeze EANN.pth -o eann_latest_"""
        return super()._resolve_freeze_command(*args, **kwargs)
    
    @property
    def frozen_name(self):
        """"""
        return f"{self.name}.pth"
    
    def write_input(self, dataset, *args, **kwargs):
        """"""

        return
    
    def read_convergence(self) -> bool:
        """"""

        return


class BeannManager(AbstractPotentialManager):

    name = "beann"
    implemented_backends = ["ase", "lammps"]

    valid_combinations = (
        # calculator, dynamics
        ("ase", "ase"), 
        ("lammps", "ase"),
        ("lammps", "lammps"),
    )

    TRAIN_INPUT_NAME = "input_nn.json"
    
    def __init__(self):

        return
    
    def _create_calculator(self, calc_params: dict) -> Calculator:
        """Create an ase calculator.

        Todo:
            In fact, uncertainty estimation has various backends as well.
        
        """
        calc_params = copy.deepcopy(calc_params)

        command = calc_params.pop("command", None)
        directory = calc_params.pop("directory", Path.cwd())
        atypes = calc_params.pop("type_list", [])

        type_map = {}
        for i, a in enumerate(atypes):
            type_map[a] = i

        # --- model files
        model_ = calc_params.get("model", [])
        if not isinstance(model_, list):
            model_ = [model_]

        models = []
        for m in model_:
            m = Path(m).resolve()
            if not m.exists():
                raise FileNotFoundError(f"Cant find model file {str(m)}")
            models.append(str(m))

        if self.calc_backend == "ase":
            # return ase calculator
            from eann.interface.ase.calculator import Eann
            if models and type_map:
                calc = Eann(
                    model=models, type_map=type_map,
                    command = command, directory=directory
                )
                calc.calc_uncertainty = True
            else:
                calc = None
        elif self.calc_backend == "lammps":
            from gdpx.computation.lammps import Lammps
            if models:
                pair_style = "eann {}".format(" ".join(models))
                # NOTE: need to specify precision
                pair_coeff = calc_params.pop("pair_coeff", "double * *")

                pair_style_name = pair_style.split()[0]
                assert pair_style_name == "eann", "Incorrect pair_style for lammps deepmd..."

                calc = Lammps(
                    command=command, directory=directory, 
                    pair_style=pair_style, pair_coeff=pair_coeff,
                    **calc_params
                )

                # - update several params
                calc.units = "metal"
                calc.atom_style = "atomic"
            else:
                calc = None

        return calc
    
    def register_calculator(self, calc_params, *args, **kwargs):
        """"""
        super().register_calculator(calc_params)

        self.calc = self._create_calculator(self.calc_params)

        return
    
    def register_trainer(self, train_params_: dict):
        """"""
        super().register_trainer(train_params_)

        return

    def _make_train_files(self, dataset=None, train_dir=Path.cwd()):
        """ make files for training
        """
        # - add dataset to config
        if not dataset:
            dataset = self.train_dataset
        assert dataset, f"No dataset has been set for the potential {self.name}."

        # TODO: for now, only List[Atoms]
        from gdpx.computation.utils import get_composition_from_atoms
        groups = {}
        for atoms in dataset:
            composition = get_composition_from_atoms(atoms)
            key = "".join([k+str(v) for k,v in composition])
            if key in groups:
                groups[key].append(atoms)
            else:
                groups[key] = [atoms]
        from ase.io import read, write
        systems = []
        dataset_dir = train_dir/"dataset"
        dataset_dir.mkdir()
        for key, frames in groups.items():
            k_dir = dataset_dir/key
            k_dir.mkdir()
            write(k_dir/"frames.xyz", frames)
            systems.append(str(k_dir.resolve()))

        dataset_config = dict(
            style = "auto",
            systems = systems,
            batchsizes = [[self.train_config["training"]["batchsize"],len(systems)]]
        )

        train_config = copy.deepcopy(self.train_config)
        train_config.update(dataset=dataset_config)

        train_config["training"]["epoch"] = self.train_epochs

        with open(train_dir/"config.yaml", "w") as fopen:
            yaml.safe_dump(train_config, fopen, indent=2)

        return
    
    def train(self, dataset=None, train_dir=Path.cwd()):
        """"""
        self._make_train_files(dataset, train_dir)
        # run command

        return
    
    def freeze(self, train_dir=Path.cwd()):
        """ freeze model and return a new calculator
            that may have a committee for uncertainty
        """
        super().freeze(train_dir)

        # - find subdirs
        train_dir = Path(train_dir)
        mdirs = []
        for p in train_dir.iterdir():
            if p.is_dir() and p.name.startswith("m"):
                mdirs.append(p.resolve())
        assert len(mdirs) == self.train_size, "Number of models does not equal model size..."

        # - find models and form committee
        models = []
        for p in mdirs:
            if self.calc_backend == "ase":
                models.append(str(p/"eann_latest_py_DOUBLE.pt"))
            elif self.calc_backend == "lammps":
                models.append(str(p/"eann_latest_lmp_DOUBLE.pt"))
        models.sort()

        # - update current calculator
        calc_params = copy.deepcopy(self.calc_params)
        calc_params["model"] = models
 
        self.calc = self._create_calculator(calc_params)

        # --- update current estimator
        # TODO: eann has interal committee in ase and lammps
        est_params = dict(
            committee = dict(
                models = models
            )
        )
        self.register_uncertainty_estimator(est_params)

        return


if __name__ == "__main__":
    pass


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
from typing import List

import numpy as np
from scipy.sparse.linalg import LinearOperator, svds
from scipy.spatial.distance import cdist

"""Methods for selection of vector-based descriptors.
"""


# -
def stat_str2val(stat: str, values: List[float]) -> float:
    """Convert a statistic string to a specific value.

    Args:
        stat: Statistics name.
        values: A List of scalar values.

    Return:
        The statistics value.

    """
    if isinstance(stat, str):
        if stat == "min":
            v = np.min(values)
        elif stat == "max":
            v = np.max(values)
        elif stat == "mean" or stat == "avg":  # Compatibilty.
            v = np.mean(values)
        elif stat == "svar":
            v = np.sqrt(np.var(values - np.average(values)))
        elif stat == "median":
            v = np.median(values)
        elif stat.startswith("percentile"):
            q = int(stat.strip().split("_")[1])  # should be within 0 and 100
            v = np.percentile(values, q)
        else:
            raise RuntimeError(f"Unknown statistics {stat}.")
    else:
        if stat == -np.inf:
            v = np.min(values)
        elif stat == np.inf:
            v = np.max(values)
        else:  # assume it is a regular number or a numpy scalar
            v = stat

    return v


# - CUR decomposition
def descriptor_svd(at_descs, num: int, do_vectors="vh"):
    """Perfrom a sparse SVD."""

    def mv(v):
        return np.dot(at_descs, v)

    def rmv(v):
        return np.dot(at_descs.T, v)

    A = LinearOperator(at_descs.shape, matvec=mv, rmatvec=rmv, matmat=mv)

    return svds(A, k=num, return_singular_vectors=do_vectors)  # sparse SVD


def cur_selection(
    features, num: int, zeta: float = 2, strategy: str = "descent", rng=np.random
):
    """Performa a CUR selection.

    Args:
        features: A (nframes,nfeatures) shaped arrray.
        num: Selected number.
        zeta: Exponential coefficient.
        strategy: Selection strategy, either stochastic or descent.
        rng: Random generator.

    References:
        [1] Bernstein, N.; Csányi, G.; Deringer, V. L.
            De Novo Exploration and Self-Guided Learning of Potential-Energy Surfaces.
            npj Comput. Mater. 2019, 5, 99.
        [2] Mahoney, M. W.; Drineas, P.
            CUR Matrix Decompositions for Improved Data Analysis.
            Proc. Natl. Acad. Sci. USA 2009, 106, 697–702.

    """
    # column vectors of descriptors
    assert features.ndim == 2
    nframes = features.shape[0]

    at_descs = features.copy().T

    # do SVD on kernel if desired
    if zeta > 0.0:
        m = np.matmul(at_descs.T, at_descs) ** zeta
    else:
        m = at_descs

    (u, s, vt) = descriptor_svd(m, min(max(1, int(num / 2)), min(m.shape) - 1))
    c_scores = np.sum(vt**2, axis=0) / vt.shape[0]

    if strategy == "stochastic":
        selected = sorted(
            rng.choice(range(nframes), size=num, replace=False, p=c_scores)
        )
    elif strategy == "descent":
        selected = sorted(np.argsort(c_scores)[-num:])
    else:
        raise ValueError("Unsupport CUR selection strategy.")

    return c_scores, selected


# - FPS (Farthest Point Sampling)
def fps_selection(
    features,
    num: int,
    min_distance=0.1,
    metric="euclidean",
    metric_params={},
    rng=np.random,
):
    """Farthest point sampling on vector-based features.

    Reference:
        https://github.com/bigd4/PyNEP.git

    """
    npoints = features.shape[0]

    selected_indices, selected_features = [], []

    # - get starting point
    start_index = rng.choice(npoints)
    selected_indices.append(start_index)
    selected_features.append(features[start_index])

    distances = np.min(
        cdist(features, selected_features, metric=metric, **metric_params), axis=1
    )  # shape (npoints,nselected) -> (npoints,)

    while np.max(distances) > min_distance or len(selected_indices) < num:
        i = np.argmax(distances)
        selected_indices.append(i)
        if len(selected_indices) >= num:
            break
        distances = np.minimum(
            distances, cdist([features[i]], features, metric=metric, **metric_params)[0]
        )  # shape (npoints,)

    scores = distances

    return scores, selected_indices


# - boltz (Boltzmann Selection)
def boltz_selection(
    boltz: float,
    props: List[float],
    input_indices: List[int],
    num_minima: int,
    rng=np.random,
):
    """Selected indices based on Boltzmann distribution.

    References:
        [1] Bernstein, N.; Csányi, G.; Deringer, V. L.
            De Novo Exploration and Self-Guided Learning of Potential-Energy Surfaces.
            npj Comput. Mater. 2019, 5, 99.

    """
    # compute desired probabilities for flattened histogram
    hist, bin_edges = np.histogram(props, bins=10)  # hits, bin_edges
    min_prop = np.min(props)

    # - multiply bin number
    config_prob = []
    for H in props:
        bin_i = np.searchsorted(bin_edges[1:], H)  # ret index of the bin
        if hist[bin_i] > 0.0:
            p = 1.0 / hist[bin_i]
        else:
            p = 0.0
        if boltz > 0.0:
            p *= np.exp(-(H - min_prop) / boltz)  # TODO: custom expression?
        config_prob.append(p)

    assert len(config_prob) == len(props)
    # uniform_probs = np.array(config_prob) / np.sum(config_prob)

    # - select
    props = copy.deepcopy(props)
    input_indices = copy.deepcopy(input_indices)

    scores, selected_indices = [], []
    for i in range(num_minima):
        # -- random
        # TODO: rewrite by mask
        config_prob = np.array(config_prob)
        config_prob /= np.sum(config_prob)
        cumul_prob = np.cumsum(config_prob)
        rv = rng.uniform()
        config_i = np.searchsorted(cumul_prob, rv)
        # print(converged_trajectories[config_i][0])
        selected_indices.append(input_indices[config_i])

        # -- remove from config_prob by converting to list
        scores.append(config_prob[config_i])
        config_prob = list(config_prob)
        del config_prob[config_i]

        # remove from other lists
        del props[config_i]
        del input_indices[config_i]

    # NOTE: scores are current probabilities when selected

    return scores, selected_indices


# - hist (Histogram-Based Selection)
def hist_selection(
    nbins: int,
    pmin: float,
    pmax: float,
    props: List[float],
    input_indices: List[int],
    num_minima: int,
    rng=np.random,
):
    props = np.array(props)
    scores = None
    selected_indices = None

    if pmin == -np.inf:
        pmin = props.min()
    if pmax == np.inf:
        pmax = props.max()
    # print("hist: ", pmin, pmax)

    bin_edges = np.linspace(pmin, pmax, nbins, endpoint=False).tolist()
    bin_edges.append(pmax)
    # print(len(bin_edges), bin_edges)

    bin_indices = np.digitize(props, bin_edges, right=False)
    # print("bin_indices: ", bin_indices)

    groups = [[] for _ in range(nbins)]
    for i, i_bin in enumerate(bin_indices):
        # dump prop not in pmin and pmax
        if 0 < i_bin <= nbins:
            groups[i_bin - 1].append(i)
    hist_by_digit = np.array([len(x) for x in groups])
    # print(hist_by_digit)

    # - select bins
    selected_groups = [[] for i in range(nbins)]

    cur_groups_ = copy.deepcopy(groups)
    for i in range(num_minima):
        # -- select bin
        cur_hists_ = np.array([len(x) for x in cur_groups_])
        curr_npoints = np.sum(cur_hists_)
        if curr_npoints > 0:
            # print("hist: ", cur_hists_)
            cur_probs_ = cur_hists_ / np.sum(cur_hists_)
            s_bin = rng.choice(nbins, 1, p=cur_probs_, replace=False)[0]
            # -- select index in the bin
            s_ind = rng.choice(cur_hists_[s_bin], 1, replace=False)[0]
            selected_groups[s_bin].append(cur_groups_[s_bin][s_ind])
            del cur_groups_[s_bin][s_ind]
        else:
            # Not enough data points for num_minima
            break
    # print([len(x) for x in selected_groups])

    # - select points
    selected_indices = []
    for x in selected_groups:
        selected_indices.extend(x)
    scores = [props[i] for i in selected_indices]

    return scores, selected_indices


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import numpy as np
import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    ...

from ase.utils.forcecurve import fit_raw

from ..core.node import AbstractNode

def get_forcefit(images):
    energies = np.array([a.get_potential_energy() for a in images])
    emin = np.min(energies)
    forces = [a.get_forces(apply_constraint=True) for a in images]
    positions = [a.positions for a in images]
    cell, pbc = images[0].get_cell(complete=True), True
    #energies -= emin
    #rxn_coords = compute_rxn_coords(images)
    #ax.scatter(rxn_coords, energies, label="dp")
    ff = fit_raw(energies, forces, positions, cell, pbc) # ForceFit
    #print(ff)

    return ff

class ReactionComparator(AbstractNode):

    def __init__(self, nimages=7, *args, **kwargs) -> None:
        """"""

        self.nimages = nimages

        return
    
    def run(self, prediction, reference):
        """"""
        # - input shape should be (2, ?, ?nimages_per_band)?
        print(f"reference: {reference}")
        print(f"prediction: {prediction}")

        reference = reference.get_marked_structures()
        prediction = prediction.get_marked_structures()

        nimages = self.nimages
        for i in range(int(len(reference)/nimages)):
            self._irun(
                prediction[i*nimages:(i+1)*nimages], 
                reference[i*nimages:(i+1)*nimages], 
                f"{i}".zfill(4)+"."
            )

        self._report()

        return

    def _irun(self, prediction, reference, prefix):
        dene = prediction[0].get_potential_energy() - reference[0].get_potential_energy()

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12, 8))
        plt.suptitle("Nudged Elastic Band Calculation")

        ax.set_xlabel("Reaction Coordinate [Å]")
        ax.set_ylabel("Potential Energy [eV]")

        self._add_axis(ax, reference, dene, "dft")
        self._add_axis(ax, prediction, 0., "dp")

        ax.legend()

        plt.savefig(self.directory/f"{prefix}neb.png")

        return
    
    def _add_axis(self, ax, images, dene: float, label):
        """"""
        ff = get_forcefit(images)
        ax.scatter(ff.path, ff.energies-dene, label=label)
        ax.plot(ff.fit_path, ff.fit_energies-dene, "k-")

        return
    
    def _report(self):
        """"""
        from reportlab.platypus import SimpleDocTemplate, Image, Paragraph

        story = []

        # - find figures
        nebfigures = list(self.directory.glob("*neb.png"))
        nebfigures = sorted(nebfigures)
        print(nebfigures)

        for i, p in enumerate(nebfigures):
            story.append(Paragraph(f"pathway {i}..."))
            image = Image(p)
            image.drawWidth = 240
            image.drawHeight = 160
            story.append(image)

        doc = SimpleDocTemplate(str(self.directory/"report.pdf"))
        doc.build(story)

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from typing import Union, List

import numpy as np

from ase import units
from ase import Atoms

from ..data.array import AtomsNDArray
from .validator import AbstractValidator

EVA2TOJNM2 = (1./(units.kJ/1000.))/((1./units.m)**2)


class SurfaceEnergyValidator(AbstractValidator):

    """This computes the surface energy of symmetric surface energies.

    TODO:
        Asymmetric surfaces.

    """

    def __init__(self, nsides=1, directory: Union[str, pathlib.Path] = "./", *args, **kwargs):
        """"""
        super().__init__(directory, *args, **kwargs)

        self.nsides = nsides

        return

    def _process_data(self, data) -> List[List[Atoms]]:
        """"""
        data = AtomsNDArray(data)

        if data.ndim == 1:
            data = [data.tolist()]
        elif data.ndim == 2: # assume it is from extract_cache...
            data = data.tolist()
        elif data.ndim == 3: # assume it is from a compute node...
            data_ = []
            for d in data[:]: # TODO: add squeeze method?
                data_.extend(d)
            data = data_
        else:
            raise RuntimeError(f"Invalid shape {data.shape}.")

        return data
    
    def run(self, dataset, worker=None, *args, **kwargs):
        """"""
        ref_grp = dataset.get("reference", None)
        if ref_grp is not None:
            ref_data = self._compute_surface_energy(
                self._process_data(ref_grp["bulk"]), 
                self._process_data(ref_grp["surfaces"]),
                self.nsides
            )
            self._write_data(ref_data, prefix="ref-")

        pre_grp = dataset.get("prediction", None)
        if pre_grp is not None:
            pre_data = self._compute_surface_energy(
                self._process_data(pre_grp["bulk"]), 
                self._process_data(pre_grp["surfaces"]),
                self.nsides
            )
            self._write_data(pre_data, prefix="pre-")

        return
    
    def _compute_surface_energy(self, bulk: List[List[Atoms]], surfaces: List[List[Atoms]], nsides=1):
        """"""
        assert len(bulk) == 1, "Only support one bulk."
        bulk_ene = bulk[0][-1].get_potential_energy()
        natoms_bulk = len(bulk[0][-1])

        data = []
        for surf in surfaces:
            # TODO: assert z-axis is perpendicular to the surface plane
            surf = [s for s in surf if s is not None]
            natoms_surf = len(surf[0])
            n_units = int(natoms_surf/natoms_bulk) # TODO: check integer units
            cell = surf[0].get_cell(complete=True)
            surf_area = np.linalg.norm(np.cross(cell[0], cell[1]))
            fin_ene = surf[-1].get_potential_energy()
            relaxed_ene = fin_ene - surf[0].get_potential_energy()
            d_ene = fin_ene - bulk_ene*n_units
            surf_ene = d_ene/(2*surf_area) + relaxed_ene/(nsides*surf_area)
            surf_ene_jnm2 = surf_ene*EVA2TOJNM2
            data.append([surf_area, d_ene, relaxed_ene, surf_ene, surf_ene_jnm2])

        return data
    
    def _write_data(self, data, prefix=""):
        """"""
        content = ("#{:>11s}  "+"{:>12s}  "*5+"\n").format(
            "N", "Area [A2]", "Delta [eV]", "RelEne [eV]", "Sigma [eV/A2]", "Sigma [J/m2]" 
        )
        for i, curr_data in enumerate(data):
            content += ("{:>12d}  "+"{:>12.4f}  "*5+"\n").format(i, *curr_data)
        self._print(f"\n{prefix}\n"+content)

        data_fpath = self.directory / f"{prefix}surfene.dat"
        with open(data_fpath, "w") as fopen:
            fopen.write(content)

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
from pathlib import Path

import numpy as np

import spglib

from ase import Atom, utils
from ase.io import read, write

"""
get adsorption sites ->
find symmetry and group structures ->
check them manually
"""
print("spglib version: ", spglib.get_version())

def find_symmetry(atoms):
    """"""
    # find symmetry
    lattice = atoms.get_cell()
    positions = atoms.get_scaled_positions()
    numbers = atoms.get_atomic_numbers()
    magmoms = None # The collinear polarizations magmoms only work with get_symmetry 
                   # and are given as a list of N floating point values
    cell = (lattice, positions, numbers)

    spg = spglib.get_spacegroup(cell, symprec=0.1)
    operations = spglib.get_symmetry(cell, symprec=0.1)
    #lattice, scaled_positions, numbers = spglib.find_primitive(cell, symprec=0.1)
    #lattice, scaled_positions, numbers = spglib.standardize_cell(cell, to_primitive=True, no_idealize=False, symprec=1e-5)
    #print("=== primitive")
    #print(lattice)
    #print(scaled_positions)
    #print(numbers)

    #return spg, operations
    return spg

# read atoms
#atoms = read("./surfaces/surf-22/O2/s1fcc1dtet-1/CONTCAR")
#atoms = read("./surfaces/surf-22/O2/surf-1fcc1hcp/CONTCAR")

surface = read("/mnt/scratch2/users/40247882/oxides/surfaces/Pt111/surf-22/surf-22/CONTCAR")
cell = surface.get_cell()

# site is on z 0.35 but O is on 0.4 after opt
#fcc_sites = [
#    [0.0, 0.0, 0.40],
#    [0.5, 0.0, 0.40],
#    [0.0, 0.5, 0.40],
#    [0.5, 0.5, 0.40]
#] # p(2x2)
fcc_sites = [
    [0.0, 0.0, 0.40],
    [0.0, 1/3., 0.40],
    [0.0, 2/3., 0.40],
    [1/3., 0.0, 0.40],
    [1/3., 1/3., 0.40],
    [1/3., 2/3., 0.40],
    [2/3., 0.0, 0.40],
    [2/3., 1/3., 0.40],
    [2/3., 2/3., 0.40]
] # p(3x3)
fcc_sites = [
    [0/4., 0.0, 0.40],
    [0/4., 1/4., 0.40],
    [0/4., 2/4., 0.40],
    [0/4., 3/4., 0.40],
    [1/4., 0.0, 0.40],
    [1/4., 1/4., 0.40],
    [1/4., 2/4., 0.40],
    [1/4., 3/4., 0.40],
    [2/4., 0.0, 0.40],
    [2/4., 1/4., 0.40],
    [2/4., 2/4., 0.40],
    [2/4., 3/4., 0.40],
    [3/4., 0.0, 0.40],
    [3/4., 1/4., 0.40],
    [3/4., 2/4., 0.40],
    [3/4., 3/4., 0.40],
] # p(4x4)
fcc_sites = np.dot(np.array(fcc_sites), cell).tolist()

hcp_sites = [
    [1./6, 1./6, 0.40],
    [2./3, 1./6, 0.40],
    [1./6, 2./3, 0.40],
    [2./3, 2./3, 0.40]
] # p(2x2)
hcp_sites = [
    [1./9, 1./9, 0.40],
    [1./9, 4./9, 0.40],
    [1./9, 7./9, 0.40],
    [4./9, 1./9, 0.40],
    [4./9, 4./9, 0.40],
    [4./9, 7./9, 0.40],
    [7./9, 1./9, 0.40],
    [7./9, 4./9, 0.40],
    [7./9, 7./9, 0.40]
] # p(3x3)
hcp_sites = [
    [1./12, 1./12, 0.40],
    [1./12, 4./12, 0.40],
    [1./12, 7./12, 0.40],
    [1./12, 10./12, 0.40],
    [4./12, 1./12, 0.40],
    [4./12, 4./12, 0.40],
    [4./12, 7./12, 0.40],
    [4./12, 10./12, 0.40],
    [7./12, 1./12, 0.40],
    [7./12, 4./12, 0.40],
    [7./12, 7./12, 0.40],
    [7./12, 10./12, 0.40],
    [10./12, 1./12, 0.40],
    [10./12, 4./12, 0.40],
    [10./12, 7./12, 0.40],
    [10./12, 10./12, 0.40],
] # p(4x4)
hcp_sites = np.dot(np.array(hcp_sites), cell).tolist()

dtet_sites = [ # down tetrahedron site
    [1./6, 1./6, 0.33],
    [2./3, 1./6, 0.33],
    [1./6, 2./3, 0.33],
    [2./3, 2./3, 0.33]
]
dtet_sites = np.dot(np.array(dtet_sites), cell).tolist()

utet_sites = [
    [1./3, 1./3, 0.30],
    [5./6, 1./3, 0.30],
    [1./3, 5./6, 0.30],
    [5./6, 5./6, 0.30]
]
utet_sites = np.dot(np.array(utet_sites), cell).tolist()

def find_sites(base_surface, num: int, sites1, sites2, site_name):
    # generate 3 O on surface
    from itertools import combinations, product
    count = 0
    for i in range(num+1):
        combins = combinations(range(len(sites1)), i)
        j = num - i
        combins2 = combinations(range(len(sites2)), j)
        cur_frames = []
        for c in product(combins, combins2):
            new_atoms = base_surface.copy()
            for x in c[0]:
                new_atoms.append(
                    Atom("O", position=sites1[x])
                )
            for x in c[1]:
                new_atoms.append(
                    Atom("O", position=sites2[x])
                )
            new_atoms.info["combination"] = c
            print("sym: ", find_symmetry(new_atoms))
            cur_frames.append(new_atoms)
            count += 1
        print(f"#frames {i}-{j}: ", len(cur_frames))
        write(site_name+"_frames_{0}-{1}.arc".format(i, j), cur_frames, format="dmol-arc")
        write(site_name+"_frames_{0}-{1}.xtd".format(i, j), cur_frames)
        #exit()
    print("number of combinations: ", count)

    return

def find_sites_symm(base_surface, num: int, sites1, sites2, site_name):
    # generate 3 O on surface
    from itertools import combinations, product
    count = 0
    for i in range(num+1):
        combins = combinations(range(len(sites1)), i)
        j = num - i
        #if i != 2 and j != 1:
        #    continue
        icount = 0
        combins2 = combinations(range(len(sites2)), j)
        cur_frames = []
        for c in product(combins, combins2):
            new_atoms = base_surface.copy()
            for x in c[0]: # site 1
                new_atoms.append(
                    Atom("C", position=sites1[x])
                )
            for x in c[1]: # site 2
                new_atoms.append(
                    Atom("O", position=sites2[x])
                )
            new_atoms.info["combination"] = c
            new_atoms = new_atoms[16:]
            # check if already have this one
            print(f"sym {icount}: ", find_symmetry(new_atoms))
            cur_frames.append(new_atoms)
            icount += 1
            count += 1
        print(f"#frames {i}-{j}: ", len(cur_frames))
        #write(site_name+"_frames_{0}-{1}.arc".format(i, j), cur_frames, format="dmol-arc")
        write(site_name+"_frames_{0}-{1}.xyz".format(i, j), cur_frames)
        #exit()
    print("number of combinations: ", count)

    return

def find_sites_on_basis(base_surface, num: int, sites1, sites2, site_name):
    # read basis structure
    import json
    basis_file = Path("./basis.json")
    if basis_file.exists():
        with open("./basis.json", "r") as fopen:
            basis_config = json.load(fopen)
        print(basis_config)
    else:
        basis_config = {}

    # generate 3 O on surface
    from itertools import combinations, product
    num_dist = [] # site number distribution
    for i in range(num+1):
        j = num - i
        cur_basis = []
        # run over basis
        for k, bc in enumerate(basis_config):
            bi, bj = len(bc["fcc"]), len(bc["hcp"])
            if i >= bi  and j >= bj:
                cur_basis.append(k)
        num_dist.append((i,j,cur_basis))
    print(num_dist)
    count = 0
    for i, j, basis in num_dist:
        if j != 0:
            print("skip hcp sites for now...")
            continue
        if len(basis) == 0:
            print("No basis structures are provided.")
            combins = combinations(range(len(sites1)), i)
            icount = 0
            combins2 = combinations(range(len(sites2)), j)
            cur_frames_sym = {}
            for c in product(combins, combins2):
                new_atoms = base_surface.copy()
                for x in c[0]: # site 1
                    new_atoms.append(
                        Atom("C", position=sites1[x])
                    )
                for x in c[1]: # site 2
                    new_atoms.append(
                        Atom("O", position=sites2[x])
                    )
                new_atoms.info["combination"] = c
                new_atoms = new_atoms[16:]
                # check if already have this one
                sym = find_symmetry(new_atoms)
                print(f"sym {icount}: ", sym)
                label = sym.split("(")[-1].split(")")[0]
                if label not in cur_frames_sym.keys():
                    cur_frames_sym[label] = [new_atoms]
                else:
                    cur_frames_sym[label].append(new_atoms)
                icount += 1
                count += 1
            for sym, cur_frames in cur_frames_sym.items():
                print(f"#frames {i}-{j} with {sym}: ", len(cur_frames))
                write(site_name+"_frames_{0}-{1}-{2}.xyz".format(i, j, "sym"+sym), cur_frames)
        else:            
            print("Use basis structures.")
            cur_frames_sym = {}
            for k in basis: # basis index
                bc = basis_config[k]

                base_atoms = base_surface.copy()
                for x in bc["fcc"]:
                    base_atoms.append(
                        Atom("C", position=sites1[x])
                    )
                for x in bc["hcp"]:
                    base_atoms.append(
                        Atom("O", position=sites2[x])
                    )

                site1_indices = [x for x in range(len(sites1)) if x not in bc["fcc"]]
                site2_indices = [x for x in range(len(sites2)) if x not in bc["hcp"]]
                combins = combinations(site1_indices, i-len(bc["fcc"]))
                combins2 = combinations(site2_indices, j-len(bc["hcp"]))

                icount = 0
                for c in product(combins, combins2):
                    new_atoms = base_atoms.copy()
                    for x in c[0]: # site 1
                        new_atoms.append(
                            Atom("C", position=sites1[x])
                        )
                    for x in c[1]: # site 2
                        new_atoms.append(
                            Atom("O", position=sites2[x])
                        )
                    new_atoms.info["combination"] = c
                    new_atoms = new_atoms[16:]
                    # check if already have this one
                    sym = find_symmetry(new_atoms)
                    print(f"sym {icount}: ", sym)
                    label = sym.split("(")[-1].split(")")[0]
                    if label not in cur_frames_sym.keys():
                        cur_frames_sym[label] = [new_atoms]
                    else:
                        cur_frames_sym[label].append(new_atoms)
                    icount += 1
                    count += 1
            for sym, cur_frames in cur_frames_sym.items():
                print(f"#frames {i}-{j} with {sym}: ", len(cur_frames))
                write(site_name+"_frames_{0}-{1}-{2}.xyz".format(i, j, "sym"+sym), cur_frames)
    print("number of combinations: ", count)

    return


site_dict = {
    "fcc": fcc_sites,
    "hcp": hcp_sites,
    "dtet": dtet_sites,
    "utet": utet_sites
}


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-n", "--number", required=True, type=int,
        help="number of O atoms added"
    )
    parser.add_argument(
        "-s", "--sites", nargs=2, required=True,
        help="two sites [fcc, hcp, dtet, utet]"
    )
    args = parser.parse_args()

    # generate configurations
    sites1, sites2 = site_dict[args.sites[0]], site_dict[args.sites[1]]
    site_name = args.sites[0] + "-" + args.sites[1]
    
    #find_sites(surface, args.number, sites1, sites2, site_name)
    #print(spglib.get_spacegroup_type("p_3_-2\""))
    #find_sites_symm(surface, args.number, sites1, sites2, site_name)
    find_sites_on_basis(surface, args.number, sites1, sites2, site_name)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import os
import subprocess

from ase.calculators.calculator import EnvironmentError, CalculationFailed


def run_ase_calculator(name: str, command, directory):
    """Run vasp from the command.

    ASE Vasp does not treat restart of a MD simulation well. Therefore, we run
    directly from the command if some input files aready exist.

    For example, we use existed INCAR for VASP.

    """
    try:
        proc = subprocess.Popen(command, shell=True, cwd=directory)
    except OSError as err:
        # Actually this may never happen with shell=True, since
        # probably the shell launches successfully.  But we soon want
        # to allow calling the subprocess directly, and then this
        # distinction (failed to launch vs failed to run) is useful.
        msg = f"Failed to execute `{command}`"
        raise EnvironmentError(msg) from err

    errorcode = proc.wait()

    if errorcode:
        path = os.path.abspath(directory)
        msg = (
            f"Calculator `{name}` failed with command `{command}` "
            + f"failed in `{path}` with error code `{errorcode}`"
        )
        raise CalculationFailed(msg)

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy

from typing import List

from itertools import combinations

import numpy as np
import scipy as sp
import networkx as nx

from ase import Atoms
from ase.io import read, write
from ase.calculators.singlepoint import SinglePointCalculator

from ..builder.group import create_an_intersect_group
from .creator import StruGraphCreator
from .comparison import bond_match, get_unique_environments_based_on_bonds
from .utils import node_symbol, unpack_node_name, show_nodes, plot_graph
from .surface import generate_normals


def make_clean_atoms(atoms_, results=None):
    """Create a clean atoms from the input."""
    atoms = Atoms(
        symbols=atoms_.get_chemical_symbols(),
        positions=atoms_.get_positions().copy(),
        cell=atoms_.get_cell().copy(),
        pbc=copy.deepcopy(atoms_.get_pbc())
    )
    if results is not None:
        spc = SinglePointCalculator(atoms, **results)
        atoms.calc = spc

    return atoms

def get_angle_cycle(a1,a2,a3):
    """Get the angle by three vectors."""
    v1 = a1-a2
    v2 = a3-a2
    #print(v1,v2)
    nv1 = np.linalg.norm(v1)
    nv2 = np.linalg.norm(v2)
    if (nv1 <= 0).any() or (nv2 <= 0).any():
        raise ZeroDivisionError('Undefined angle')
    v1 /= nv1
    v2 /= nv2
    #print(v1,v2)
    cosang = np.dot(v1, v2)
    sinang = np.linalg.norm(np.cross(v1, v2))

    return np.arctan2(sinang, cosang)

def plane_normal(xyz):
    """Return the surface normal vector to a plane of best fit. THIS CODE IS BORROWED FROM CATKIT

    Parameters
    ----------
    xyz : ndarray (n, 3)
        3D points to fit plane to.

    Returns
    -------
    vec : ndarray (1, 3)
        Unit vector normal to the plane of best fit.
    """
    A = np.c_[xyz[:, 0], xyz[:, 1], np.ones(xyz.shape[0])]
    vec, _, _, _ = sp.linalg.lstsq(A, xyz[:, 2])
    vec[2] = -1.0

    vec /= -np.linalg.norm(vec)

    return vec

def normalize(vector):
    return vector / np.linalg.norm(vector) if np.linalg.norm(vector) != 0 else vector * 0

class SingleAdsorptionSite(object):

    #: Minimum inter-adsorbate distance.
    MIN_INTERADSORBATE_DISTANCE: float = 1.5

    """A site for single-dentate adsorption.
    """

    def __init__(
        self, atoms: Atoms, nl, normals,
        site_indices: List[int], node_names: List[str], surf_indices: List[int],
        graph=None, *args, **kwargs
    ):
        """

        Args:
            site_indices: Indices of site atoms.

        """
        self.atoms = make_clean_atoms(atoms)

        # print(offsets)
        # assert len(cycle) == len(offsets), "site atomic number is inconsistent..."
        self.site_indices = site_indices
        # self.offsets = offsets
        # self.known = known

        order = len(self.site_indices)
        self.order = order

        self.position, self.normal, self.tangent = self._compute_position(
            atoms, nl, normals, node_names, order, site_indices, surf_indices
        )

        self.graph = graph

        return

    def __eq__(self, other):
        return nx.is_isomorphic(self.graph, other.graph, edge_match=bond_match)

    def __repr__(self):
        content = (
            "Cycle:{}, " + 
            "Position: [" + "{:<8.4f}, "*3 +"], " + 
            "Normal: [" + "{:8.4f}, "*3 + "], " +
            "Graph: {}"
        ).format(self.site_indices, *self.position, *self.normal, bool(self.graph))
        return content

    def _compute_position(
        self, atoms: Atoms, nl, normals, node_names: List[str], order: int,
        site_indices: List[int], surf_indices: List[int]
    ):
        """Compute the position of the site based on the structure graph.

        Args:
            atoms: Structure.
            order: Coordination numebr of the site (top 1, bridge 2, ...).
        
        Returns:
            Site positions and its plane normal.

        """
        assert len(site_indices) == order, "Site order is inconsistent."

        # - find positions of atoms forming the site and the normal
        # print("xxx")
        # print(names)
        # calc centre
        site_positions = np.zeros((order,3))
        for ci, n in enumerate(node_names):
            chem_sym, a_idx, offset = unpack_node_name(n)
            site_positions[ci] = atoms.positions[a_idx] + np.dot(offset, atoms.get_cell())
            # print(cycle[ci], site_positions[ci], offset)
        average = np.average(site_positions, axis=0)
        # print("site pos: ", site_positions)

        # generate site position
        """
        # TODO: is it okay for order 4 site?
        tracked = np.array(atoms[cycle[0]].position, dtype=float)
        known = np.zeros(shape=(coordination, 3), dtype=float)
        known[0] = tracked
        print(known)
        for index, (start, end) in enumerate(zip(cycle[:-1], cycle[1:])):
            nei_indices, offsets = nl.get_neighbors(start)
            print("start: ", start, end, nei_indices)
            for neighbor, offset in zip(nei_indices, offsets):
                if neighbor == end:
                    tracked += offset_position(atoms, neighbor, offset) - atoms[start].position
                    known[index + 1] = tracked

        average = np.average(known, axis=0)
        """

        # --- NOTE: add default distance to the site
        # if coordination == 2:
        #    average[2] = average[2] + self.shift2
        # if coordination == 3:
        #    average[2] = average[2] + self.shift3

        # --- generate site normal --- TODO: ?
        normal = np.zeros(3)

        if order == 1:
            for index in site_indices:
                neighbors = len(nl.get_neighbors(index)[0])
                normal += normals[index] * (1/neighbors)
            normal = normalize(normal)
        else:
            if order > 1: # TODO?
                index = order - 1
                neighbors = len(nl.get_neighbors(index)[0])
                cycle_orig = order
                # print(cycle)
                normal = compute_site_normal(atoms, nl, site_indices, surf_indices)
                # print(cycle,normal)
            else:
                # NOTE: order < 1 is not allowed
                ...

        for index in site_indices:
            neighbors = len(nl.get_neighbors(index)[0])
            normal += normals[index] * (1/neighbors)
        normal = normalize(normal)

        # - generate site tangent
        tangent = np.zeros(3)
        if order == 1:
            ...
        elif order == 2:
            tangent = site_positions[1] - site_positions[0]
        elif order == 3:
            tangent = site_positions[2] - (site_positions[0]+site_positions[1])/2.
        else:
            ...

        return average, normal, tangent

    def is_occupied(self, ads_elemnts):
        """"""
        ads_indices = [a.index for a in self.atoms if a.symbol in ads_elemnts]
        ads_nodes = None
        ads_nodes = [node_symbol(self.atoms[i], (0, 0, 0)) for i in ads_indices]

        ads_graphs = nx.subgraph(self.graph, ads_nodes)
        ads_graphs = [ads_graphs.subgraph(c) for c in nx.connected_components(ads_graphs)]
        print("number of adsorbate graphs: ", len(ads_graphs))
        # print(self.offsets)

        for idx, ads in enumerate(ads_graphs):
            print(f"----- adsorbate {idx} -----")
            print("ads nodes: ", ads.nodes())
            initial = list(ads.nodes())[0] # the first node in the adsorbate
            full_ads = nx.ego_graph(self.graph, initial, radius=0, distance="ads_only") # all nodes in this adsorbate, equal ads?
            # print("full ads: ", full_ads.nodes())

            new_ads = nx.ego_graph(self.graph, initial, radius=1, distance="dist") # consider neighbour atoms
            new_ads = nx.Graph(nx.subgraph(self.graph, list(new_ads.nodes()))) # return a new copy of graph

            # update attribute of this adsorbate
            for node in ads.nodes():
                new_ads.add_node(node, central_ads=True) # central means the adsorbate interacting with surface

            # update attr of this and neighbour adsorbates
            for node in full_ads.nodes():
                new_ads.add_node(node, ads=True)

            show_nodes(new_ads)

        return

    def adsorb(
        self, 
        adsorbate_, # single atom or a molecule
        other_ads_indices, # indices of other adsorbate
        ads_params: List[dict],
        check_H_bond=False
    ) -> List[Atoms]:
        """Adsorb atoms/molecule on the substrate.

        Several adsorption modes are supported.

        Args:
            ads_params: Adsorption-related parameters.

        """
        # - find indices of adsorbates
        #    NOTE: not check distance to hydrogens since
        #          since the distances are usually very small
        natoms, natoms_ads = len(self.atoms), len(adsorbate_)
        inserted_ads_indices = [i+natoms for i in range(natoms_ads) if adsorbate_[i].symbol != "H"]

        # adsorbates that are already in the substrate
        local_ads_indices = [i for i in other_ads_indices if self.atoms[i].symbol != "H"]

        # TODO: assert adsorbate to insert is a single atom or a molecule
        # -- for planar molecules such as CO2
        # adsorbate.rotate([0, 1, 0], self.normal, center=[0,0,0])
        # adsorbate.translate(self.position + (self.normal*distance_to_site))

        # -- for complex molecules

        # - start to adsorb
        ads_frames = []
        for cur_params in ads_params:
            # -- params
            mode = cur_params.get("mode", "atop")
            distance_to_site = cur_params.get("distance", 1.5)
            # -- prepare and add adsorbate
            substrate = self.atoms.copy()
            adsorbate = adsorbate_.copy()
            com = adsorbate.get_center_of_mass()
            adsorbate.translate(np.zeros(3)-com) # first translate to origin
            # --
            if mode == "atop":
                # -- for single atom or linear molecule
                adsorbate.rotate([0, 0, 1], self.normal, center=[0,0,0])
                adsorbate.translate(self.position + (self.normal*distance_to_site))
                substrate.extend(adsorbate)
            # -- more adsoprtion modes
            elif mode == "tbt": # parallel to the site
                assert self.order == 2, "Only the bridge site has the tbt adsorption."
                # NOTE: only for bridge and hollow sites
                #       default atom/molecule is along z-axis
                adsorbate.rotate([0, 0, 1], self.tangent, center=[0,0,0])
                adsorbate.translate(self.position + (self.normal*distance_to_site))
                substrate.extend(adsorbate)
            elif mode == "thb":
                assert self.order == 3, "Only the hollow site has the thb adsorption."
                adsorbate.rotate([0, 0, 1], self.tangent, center=[0,0,0])
                adsorbate.translate(self.position + (self.normal*distance_to_site))
                substrate.extend(adsorbate)
            else:
                raise NotImplementedError(f"Adsorption mode {mode} is not supported.")

            # - check distances between adsorbates
            dist = float("inf")
            if len(local_ads_indices) != 0:
                for i in inserted_ads_indices:
                    dists = substrate.get_distances(i, local_ads_indices, mic=True)
                    dist = min(dist, dists.min())
                # TODO: check is_occupied
                if dist < self.MIN_INTERADSORBATE_DISTANCE:
                    substrate = None
            if substrate is not None: 
                ads_frames.append(substrate)

        return ads_frames

    def _iadsorb(self):
        """"""

        return


def compute_site_normal(atoms: Atoms, nl, site_indices: List[int], surf_indices: List[int]):
    """Compute the plane normal of the site.

    Atoms that have positive z-axis normal would be on the surface.

    Args:
        atoms: Atoms object.
        nl: Pre-created neighbour list.
        site_indices: Indices of atoms that form the adsorption site.
        surf_indices: Indices of surface atoms.
    
    Returns:
        An array with shape (3,).

    """
    site_indices = copy.deepcopy(site_indices)
    site_order = len(site_indices)
    if site_order > 2:
        #atom_array = []
        #for a in site_indices:
        #    atom_array.append(atoms[a].position)
        positions = atoms.positions[site_indices,:]
        normal = plane_normal(positions) # TODO?
    else: # for bridge site
        neighbor_atoms = []
        #print(self.nl.get_neighbors(site_atoms[0])[0])
        #print(self.nl.get_neighbors(site_atoms[1])[0])
        for i in nl.get_neighbors(site_indices[0])[0]:
            #NOTE: This if condition is to ensure that if the atoms are the same row, 
            # then the plane is formed btwn an atom in another row
            # print(i,atoms[i].position[2] - atoms[site_atoms[0]].position[2])
            if (i not in site_indices) and i in nl.get_neighbors(site_indices[1])[0] and i in surf_indices and i not in neighbor_atoms:
                neighbor_atoms.append(i)
        #print("neighs: ", neighbor_atoms)

        normal = [0, 0, 0]
        if len(neighbor_atoms) > 0:
            for i in neighbor_atoms:
                site_atoms1 = site_indices.copy()
                site_atoms1.append(i)
                atom_array = []
                initial = site_indices[0]
                atom_array.append(atoms[initial].position)
                for a in site_atoms1:
                    if a != initial:
                        a_offset = nl.get_neighbors(initial)[1][np.where(a==nl.get_neighbors(initial)[0])]
                        #print(a,np.dot(a_offset, atoms.get_cell())+atoms[a].position)
                        atom_array.append(atoms[a].position+np.dot(a_offset, atoms.get_cell())[0])
                #print(atom_array)
                #print(get_angle_cycle(*atom_array))
                # NOTE: angle settings
                if 0.85 < get_angle_cycle(atom_array[0],atom_array[1],atom_array[2]) < 1.2: # 48.73 to 68.79 degree
                    normal += plane_normal(np.array(atom_array))
                #normal += plane_normal(np.array(atom_array))
                    #print('using this cycle to add to normal')
                #print(normal)
        # check normal
        if np.sum(np.fabs(normal)) < 1e-8:
            initial = site_indices[0]
            atom_array = []
            atom_array.append(atoms[initial].position)
            for a in site_indices:
                if a != initial:
                    a_offset = nl.get_neighbors(initial)[1][np.where(a==nl.get_neighbors(initial)[0])]
                    #print(a,np.dot(a_offset, atoms.get_cell())+atoms[a].position)
                    atom_array.append(atoms[a].position+np.dot(a_offset, atoms.get_cell())[0])
            vec = atom_array[1] - atom_array[0]
            nvec = vec / np.linalg.norm(vec)
            normal = nvec + np.array([0.,0.,2**0.5])

        #print("func normal: ", normal)
    normal = normalize(normal)

    #print("func normal: ", normal)

    return normal

def find_valid_indices(atoms, species: None, spec_indices=None, region=None):
    """Find indices of atoms based on criteria."""
    chemical_symbols = atoms.get_chemical_symbols()

    if spec_indices is None:
        if region is None:
            valid_indices = [] # selected indices
            for i, sym in enumerate(chemical_symbols):
                if sym == species:
                    valid_indices.append(i)
        else:
            valid_indices = []
            #print(region)
            (ox, oy, oz, xl, yl, zl, xh, yh, zh) = region
            for i, a in enumerate(atoms):
                if a.symbol == species:
                    pos = a.position
                    if (
                        (ox+xl <= pos[0] <= ox+xh) and
                        (oy+yl <= pos[1] <= oy+yh) and
                        (oz+zl <= pos[2] <= oz+zh)
                    ):
                        valid_indices.append(i)
    else:
        valid_indices = copy.deepcopy(spec_indices)
    
    return valid_indices


class SiteFinder(StruGraphCreator):

    """Find adsorption sites on surface.

    The procedure includes several steps: 1. Detect surface atoms. 2. Find site 
    with different orders (coordination number). This supports both mono- and 
    multi-dentate sites.

    """

    #: The minimum norm to be considered as surface.
    surface_normal: float = 0.65

    _site_radius = 2

    # bind an atoms object
    atoms: Atoms = None

    nl = None
    graph: nx.Graph = None

    def __init__(self, site_radius=2, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)
        # TODO: set params...
        #pbc_grid = kwargs.get("pbc_grid", [2,2,0])
        #graph_radius = kwargs.get("graph_radius", 2)
        #self.neigh_creator = NeighGraphCreator(
        #    covalent_ratio=1.0, skin=0.0, self_interaction=False, bothways=True, 
        #    graph_radius=graph_radius, pbc_grid=pbc_grid, *args, **kwargs
        #)
        self._site_radius = site_radius

        return
    
    @property
    def site_radius(self):

        return self._site_radius
    
    def find(self, atoms_: Atoms, site_params: list, surface_mask: List[int]=None):
        """Find all sites based on graph.

        The site finding includes several steps. 1. Create a graph representation of 
        the structure. 2. Determine the surface normal. 3. Determine valid sites.

        Args:
            site_params: Parameters that define the site to find.

        """
        # - input atoms
        atoms = atoms_

        # - find adsorbate atoms' indices
        ads_indices = [a.index for a in atoms if a.symbol in self.adsorbate_elements]

        # - create graph
        graph = self.generate_graph(atoms, ads_indices) # create self.graph

        # NOTE: update nl since we want use nl with bothways instead of false
        nl = self.neigh_creator.build_neighlist(atoms, bothways=True)
        nl.update(atoms)

        # TODO: surface region group command?
        # - find atoms that are on surface and possibly form adsorption sites...
        # NOTE: make sure to manually set the normals for 2-D materials, 
        #       all atoms should have a normal pointing up, as all atoms are surface atoms
        system = self.check_system(atoms)

        self._print(f"{atoms.pbc = }")
        self._print(f"{system = }")

        normals, surface_mask = generate_normals(
            atoms, nl, surface_mask, 
            ads_indices, mask_elements=[],
            surf_norm_min=self.surface_normal, normalised=True,
            system=system
        )
        self._print(f"number of surface atoms: {len(surface_mask)}")
        self._debug(f"{surface_mask = }")
        self._debug(f"surface mask ads: {ads_indices}")

        atoms.arrays["surface_direction"] = normals

        # - find all valid adsorption sites
        site_groups = []
        for i, curr_site_params in enumerate(site_params):
            self._print(f"site type {i}: {curr_site_params}")
            curr_params_list = self._broadcast_site_params(curr_site_params)
            #print(cur_params_list)

            nanchors = len(curr_params_list)

            if nanchors == 1:
                self._print("Monodentate sites...")
                curr_params = curr_params_list[0]
                sites = self._generate_single_site(
                    curr_params, atoms, graph, nl, surface_mask, normals
                )
                ...
            else:
                self._print("Multidentate sites...")
                #for curr_params in curr_params_list:
                #    ...
                raise NotImplementedError()
            
            site_groups.append(sites)

        return site_groups
    
    def _generate_single_site(
        self, params: dict, atoms: Atoms, graph, nl, surface_mask, normals
    ):
        """Generate single-multidentate site."""
        # - some basic params
        cn = params.get("cn") # coordination number
        cur_site_radius = params.get("radius", self._site_radius)
        self._print(f"coordination number: {cn}")

        # - find possible atoms to form the site
        self._print(f"graph: {graph}")
        #cur_species = params.get("species", None)
        #site_indices = params.get("site_indices", None)
        #region = params.get("region", None)
        #cur_surf_mask = find_valid_indices(
        #    atoms, cur_species, spec_indices=site_indices, region=region
        #)
        natoms = len(atoms)
        group_commands = params.get(
            "group", 
            # default: return indices of all atoms
            ["id {}".format(" ".join([str(i) for i in range(1,natoms+1)]))] # start from 1
        )
        valid_indices = create_an_intersect_group(atoms, group_commands)

        # -- Site indices must be in the surface indices
        valid_indices = [i for i in valid_indices if i in surface_mask]

        self._print(f"number of site atoms: {len(valid_indices)}")
        self._debug(f"atomic indices for site: {valid_indices}")

        # -- create sites
        found_sites = self._generate_site(
            atoms, graph, nl,
            valid_indices, surface_mask, normals, coordination=cn
        )

        # - generate site graph TODO: move this to site object?
        for s in found_sites:
            s.graph = self.process_site(
                atoms, graph, nl, s.site_indices, cur_site_radius
            )
        #print(found_sites)
        
        if found_sites:
            # -- get sites with unique environments
            #print("\n\nfound sites: ", len(found_sites))
            #for s in found_sites:
            #    print(s, s.graph)
            site_graphs = [s.graph for s in found_sites]
            unique_indices = get_unique_environments_based_on_bonds(site_graphs)
            #print(unique_indices)
            #print(unique_indices)
            unique_sites = [found_sites[i] for i in unique_indices]
        
            # TODO: sort sites by positions?
            self._print(f"unique sites: {len(found_sites)} -> {len(unique_sites)}")
        else:
            unique_sites = []
            self._print("No sites are found.")

        return unique_sites
    
    def _broadcast_site_params(self, params: dict) -> List[dict]:
        """"""
        # params need to broadcast
        keys = ["cn", "radius", "group", "distance"]
        nums = []
        for key in keys:
            cur_num = 1
            if key == "cn": # int
                # coordination number of the site
                value = params.get(key, 1)
                if isinstance(value, list):
                    cur_num = len(value)
            elif key == "radius": # int
                value = params.get(key, 3)
                if isinstance(value, list):
                    cur_num = len(value)
                    num = cur_num
            elif key == "group": # List[str]
                # NOTE: this param is not broadcastable
                #value = params.get(key, None)
                #if isinstance(value, list): # all atoms
                #    cur_num = len(value)
                #    num = cur_num
                ...
            elif key == "distance": # float
                value = params.get(key, 1.5)
                if isinstance(value, list):
                    cur_num = len(value)
                    num = cur_num
            else:
                ...
            nums.append(cur_num)

        # number of sites for a single adsorption
        num = max(nums) # consider a single-dentate site as default if num == 1 
        for key, cur_num in zip(keys, nums):
            assert (cur_num == 1 or cur_num == num), f"Parameter {key} size is inconsistent."
        
        # - create new parameters
        params_list = []
        for i in range(num):
            cur_params = dict()
            for key in keys:
                if key == "cn":
                    # coordination number of the site
                    value = params.get(key, 1)
                    if isinstance(value, list):
                        cur_params[key] = value[i]
                    else:
                        cur_params[key] = value
                elif key == "radius":
                    value = params.get(key, 3)
                    if isinstance(value, list):
                        cur_params[key] = value[i]
                    else:
                        cur_params[key] = value
                elif key == "group":
                    # NOTE: this param is not broadcastable
                    value = params.get(key, None)
                    #if isinstance(value, list): # all atoms
                    #    cur_num = len(value)
                    #    num = cur_num
                    cur_params[key] = value
                elif key == "distance":
                    value = params.get(key, 1.5)
                    if isinstance(value, list):
                        cur_params[key] = value[i]
                    else:
                        cur_params[key] = value
                else:
                    ...
            params_list.append(cur_params)

        return params_list
 
    def _generate_site(
        self, 
        atoms: Atoms,
        graph,
        nl,
        valid_surface_mask,
        surface_mask, # indices of surface atoms that can form sites
        normals, 
        coordination: int
    ) -> List[SingleAdsorptionSite]:
        """Create adsorption sites.

        Args:
            coordination: Coordination number (order) of the site.
        """
        # NOTE: check if is None
        possible = list(combinations(set(valid_surface_mask), coordination))

        # check if the pair is neighbour
        valid = []
        for cycle in possible:
           for start, end in combinations(cycle, 2):
               if end not in nl.get_neighbors(start)[0]:
                   break
           else: # All were valid
                valid.append(list(cycle))
        self._debug(f"DEBUG: {valid}")

        #print(valid)
        sites = []
        for cycle in valid:
            # - find nodes form the site
            #   NOTE: the same atoms may form different sites due to PBC
            #         if the cell is very small...
            site_node_names = []
            for u, d in graph.nodes.data():
                #print(u, d)
                if d["index"] in cycle:
                    site_node_names.append(u)
            site_graph = nx.subgraph(graph, site_node_names)
            #site_graphs = [site_graph.subgraph(c) for c in nx.connected_components(site_graph)]
            #for i, x in enumerate(site_graphs):
            #    print(f"--- {i} ---")
            #    print(x)
            #plot_graph(site_graph, "sg.png")
            centre_node = node_symbol(atoms[cycle[0]].symbol, cycle[0], (0,0,0))
            #print("centre: ", centre_node)
            local_graph = nx.ego_graph(
                site_graph, centre_node, 
                radius=self.DIS_SURF2SURF, center=True, # 2 is surf_surf_dis
                distance="dist"
            ) # go over framework
            #plot_graph(local_graph, f"local-{coordination}.png")

            site_node_names = [] # List[List[str]]
            if coordination == 3:
                #print("node names: ", local_graph.nodes())
                all_cliques = nx.enumerate_all_cliques(local_graph)
                site_node_names = [x for x in all_cliques if len(x)==3]
            # TODO: add CN4 site that is a square
            else:
                # NOTE: for very small unit cell, 
                #       nodes maybe connected by their neighbour images
                #       e.g. bridge sites on p(2x2)-Pt(111)
                nodes_comb = combinations(local_graph.nodes, coordination)
                new_site_graphs = []
                for node_names in nodes_comb:
                    if coordination == 2:
                        # check they are connected
                        if graph.has_edge(node_names[0],node_names[1]):
                            new_site_graphs.append(nx.subgraph(graph, node_names))
                    else:
                        # top site
                        new_site_graphs.append(nx.subgraph(graph, node_names))

                # *** check if node number is valid
                for g in new_site_graphs:
                    node_names = list(g.nodes())
                    assert len(node_names) == coordination, "The site order is incorrect."
                    site_node_names.append(node_names)

            # --- create site object ---
            #self.pfunc(site_node_names)
            for node_names in site_node_names:
                site_ads = SingleAdsorptionSite(
                    atoms=atoms, nl=nl, normals=normals, site_indices=cycle,
                    node_names=node_names, surf_indices=surface_mask,
                )
                sites.append(site_ads)

        return sites
    
    def process_site(
        self, 
        atoms,
        graph,
        nl,
        site: List[int],
        radius=3
    ):
        """Create the graph representation of the site.

        Args:
            site: Indices of atoms that form the site.
        
        Steps:
            1. add a placeholder X node
            2. find all neighbours of site atoms
            this can include built-in adsorbates
        """
        # - add a placeholder X node in the box [0,0,0]
        #print("radius: ", radius)
        full = graph
        #neighbors = nl.neighbors
        #offsets = nl.displacements
        #neighbors, offsets = nl.get_neighbors()
        full.add_node("X", index=None)
        offset = np.array([0, 0, 0])
        full.add_edge(
            "X",
            node_symbol(atoms[site[0]].symbol, site[0], offset),
            bond="X:{}".format(atoms[site[0]].symbol),
            ads=0
        ) # Handle first manually

        for last_index, next_index in zip(site[:-1], site[1:]):
            # Error handling needed, .index could be None / -1?
            neighbors, offsets = nl.get_neighbors(last_index)
            #neighbor_index = list(neighbors[last_index]).index(next_index)
            neighbor_index = list(neighbors).index(next_index)
            #offset += offsets[last_index][neighbor_index]
            offset += offsets[neighbor_index]
            #print(offset)
            full.add_edge(
                "X",
                node_symbol(atoms[next_index].symbol, next_index, offset),
                bond="X:{}".format(atoms[next_index].symbol),
                ads=0
            )

        # TODO: ads also in site_graph?
        site_graph = nx.ego_graph(full, "X", radius=(radius*2)+1, distance="dist")
        site_graph = nx.subgraph(full, list(site_graph.nodes()))
        site_graph = nx.Graph(site_graph)
        full.remove_node("X")

        return site_graph


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
import pathlib
from typing import Union, List, NoReturn

import omegaconf

from ase import Atoms
from ase.io import read, write

from ..core.variable import Variable
from ..core.operation import Operation
from ..core.register import registers

from ..worker.worker import AbstractWorker
from ..data.array import AtomsNDArray

from ..builder.interface import build, BuilderVariable

from .selector import AbstractSelector, load_cache
from .composition import ComposedSelector


@registers.variable.register
class SelectorVariable(Variable):

    def __init__(self, selection: Union[dict, List[dict]], directory="./", *args, **kwargs) -> None:
        """Define a Variable that has a Selector."""
        # We can define a selector in two different ways:
        # The Dict must have a selection key
        # - a Dict that defines a single selector
        # - a List of Dict that defines several selectors, 
        #   which will be converted into a composed one
        selection = copy.deepcopy(selection)
        if isinstance(selection, dict) or isinstance(selection, omegaconf.dictconfig.DictConfig):
            selection = [selection]
        elif isinstance(selection, list) or isinstance(selection, omegaconf.listconfig.ListConfig):
            ...
        else:
            raise TypeError(f"Unknown type of {selection =}.")

        selectors = []
        for params in selection:
            method = params.pop("method", None)
            selector = registers.create("selector", method, convert_name=True, **params)
            selectors.append(selector)
        nselectors = len(selectors)
        if nselectors > 1:
            selector = ComposedSelector(selectors)
        else:
            selector = selectors[0]

        super().__init__(initial_value=selector, directory=directory)

        return


@registers.operation.register
class select(Operation):

    cache_fname = "selected_frames.xyz"

    def __init__(self, structures, selector, directory="./", *args, **kwargs):
        """"""
        super().__init__(input_nodes=[structures, selector], directory=directory)

        return

    @Operation.directory.setter
    def directory(self, directory_) -> NoReturn:
        """"""
        super(select, select).directory.__set__(self, directory_)

        return

    def _preprocess_input_nodes(self, input_nodes):
        """"""
        structures, selector = input_nodes
        if isinstance(structures, str) or isinstance(structures, pathlib.Path):
            # TODO: check if it is a molecule name
            structures = build(
                BuilderVariable(
                    directory=self.directory / "structures",
                    method="reader",
                    fname=structures,
                )
            )
        # We can define a selector in two different ways:
        # The Dict must have a selection key
        # - a Dict that defines a single selector
        # - a List of Dict that defines several selectors, 
        #   which will be converted into a composed one
        if isinstance(selector, dict) or isinstance(
            selector, omegaconf.dictconfig.DictConfig
        ):
            selector = SelectorVariable(
                directory=self.directory / "selector", **selector
            )
        #self._print(f"{selector = }")

        return structures, selector

    def forward(
        self, structures: AtomsNDArray, selector: AbstractSelector
    ) -> AtomsNDArray:
        """"""
        super().forward()
        selector.directory = self.directory

        cache_fpath = self.directory / self.cache_fname
        if not cache_fpath.exists():
            new_frames = selector.select(structures)
            write(cache_fpath, new_frames)
        else:
            markers = load_cache(selector.info_fpath)
            structures.markers = markers
            if cache_fpath.stat().st_size != 0:
                new_frames = read(cache_fpath, ":")
            else: # sometimes selection gives no structures and writes empty file
                new_frames = []
        self._print(f"nframes: {len(new_frames)}")

        num_new_frames = len(new_frames)
        if num_new_frames > 0:
            self.status = "finished"
        else:
            self.status = "exit"

        return structures
    
    # NOTE: This operation exits when no structures are selected
    #       so we donot need convergence check here?
    #def report_convergence(self, *args, **kwargs) -> bool:
    #    """"""
    #    input_nodes = self.input_nodes
    #    assert self.status == "finished", f"Operation {self.directory.name} cannot report convergence without forwarding."
    #    selector = input_nodes[1].output

    #    self._print(f"{selector.__class__.__name__} Convergence")

    #    cache_fpath = self.directory / self.cache_fname # MUST EXIST
    #    if cache_fpath.stat().st_size != 0:
    #        new_frames = read(cache_fpath, ":")
    #    else: # sometimes selection gives no structures and writes empty file
    #        new_frames = []
    #    num_new_frames = len(new_frames)
    #    if num_new_frames == 0:
    #        converged = True
    #    else:
    #        converged = False

    #    return converged


def run_selection(
    param_file: Union[str, pathlib.Path],
    structure: Union[str, dict],
    directory: Union[str, pathlib.Path] = "./",
) -> None:
    """Run selection with input selector and input structures.

    This no more accepts a worker as all data used in the selection should be 
    computed in advance.
    
    """
    directory = pathlib.Path(directory)
    if not directory.exists():
        directory.mkdir(parents=True, exist_ok=False)

    from gdpx.utils.command import parse_input_file

    params = parse_input_file(param_file)

    selector = SelectorVariable(
        directory=directory, **params
    ).value
    selector.directory = directory

    # - read structures
    from gdpx.builder import create_builder

    builder = create_builder(structure)
    frames = builder.run()  # -> List[Atoms]

    # TODO: convert to a bundle of atoms?
    data = AtomsNDArray(frames)

    # -
    selected_frames = selector.select(data)

    from ase.io import read, write

    write(directory / "selected_frames.xyz", selected_frames)

    return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import List
from gdpx.potential.trainer import AbstractTrainer


class FGPTrainer(AbstractTrainer):

    name = "fgp"

    def __init__(
        self, config: dict, type_list: List[str] = None, train_epochs: int = 200, 
        directory=".", command="train", freeze_command="freeze", random_seed: int = None, 
        *args, **kwargs
    ) -> None:
        """"""
        super().__init__(
            config, type_list, train_epochs, 
            directory, command, freeze_command, 
            random_seed, *args, **kwargs
        )

        return

    @property
    def frozen_name(self):
        """"""
        return f"{self.name}.pb"

    def _resolve_freeze_command(self, *args, **kwargs):
        return super()._resolve_freeze_command(*args, **kwargs)

    def _resolve_train_command(self, *args, **kwargs):
        """"""
        command = self.command

        return command
    
    def train(self, dataset, init_model=None, *args, **kwargs):
        """"""
        self._print("miaow")
        from .representation import train
        train()

        return
    
    def write_input(self, dataset, *args, **kwargs):
        return super().write_input(dataset, *args, **kwargs)

    def read_convergence(self) -> bool:
        return super().read_convergence()


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import functools
import pathlib

import numpy as np

import jax
import jax.numpy as jnp

from ase.calculators.calculator import Calculator


@jax.jit
def gaussian_bias(x, x_t, sigma, omega):
    """Compute bias potential.

    Args:
        x: (1, num_dim)
        x_t: (num_bias, num_dim)
        sigma: (1, num_dim)
        omega: scalar

    """
    x2 = (x - x_t)**2/2./sigma**2
    v = omega*jnp.exp(-jnp.sum(x2, axis=1))

    return v.sum(axis=0)


@functools.partial(jax.jit, static_argnums=(1,))
def compute_bias(positions, cvfunc, cvparams, ref_colvars, sigma, omega):
    """"""
    colvar = cvfunc(positions, cvparams)

    bias = gaussian_bias(colvar, ref_colvars, sigma, omega)

    return bias, colvar


class GaussianCalculator(Calculator):

    implemented_properties = ["energy", "forces"]

    default_parameters = dict(
        width = 0.1,
        height = 0.2,
    )

    _history = []

    _fname = "HILLS"

    def __init__(self, colvar, restart=None, label=None, atoms=None, directory=".", **kwargs):
        """"""
        super().__init__(restart=restart, label=label, atoms=atoms, directory=directory, **kwargs)

        self._colvar = colvar
        print(self._colvar)

        _width = np.array(self.parameters["width"])
        if len(_width.shape) == 0:
            _width = _width.reshape(-1)
        self._width = _width
        self._height = np.array(self.parameters["height"])

        return
    
    def _save_history(self, ):
        """"""

        return
    
    def calculate(self, atoms=None, properties=["energy"], system_changes=["positions","numbers","cell"]):
        """"""
        super().calculate(atoms, properties, system_changes)

        sigma, omega = self._width, self._height
        
        positions = atoms.positions
        cvfunc, cvparams = self._colvar.cvfunc, self._colvar.params

        if False: # Combined function...
            num_his = len(self._history)
            if num_his == 0:
                curr_colvar = cvfunc(positions, cvparams)
                natoms = len(atoms)
                e, f = omega, np.zeros((natoms, 3))
            else:
                ref_colvars = np.vstack(self._history)
                (e, curr_colvar), f = jax.value_and_grad(compute_bias, argnums=0, has_aux=True)(
                    positions, cvfunc, cvparams, ref_colvars, sigma, omega
                )
            self._history.append(curr_colvar)
        else:
            curr_colvar = cvfunc(atoms, cvparams)
            print(f"colvar: {curr_colvar}")
            ...

        # ---
        fpath = pathlib.Path(self.directory)/self._fname
        fmode = "w"
        if fpath.exists():
            fmode = "a"
        
        with open(fpath, fmode) as fopen:
            content = f"{curr_colvar[0][0]:<8.4f}  {sigma[0]:<8.4f}  {omega:<8.4f}\n"
            fopen.write(content)

        self.results["energy"] = np.asarray(e)
        self.results["forces"] = np.array(f)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import itertools
import dataclasses
import os
import pathlib
import shutil
import traceback

from typing import Union, List

import numpy as np

from ase import Atoms
from ase import units
from ase.io import read, write
from ase.calculators.cp2k import parse_input, InputSection
from ase.calculators.singlepoint import SinglePointCalculator
from ase.neb import NEB

from .string import AbstractStringReactor, StringReactorSetting
from .. import parse_constraint_info


def run_cp2k(name, command, directory):
    """Run vasp from the command. 
    
    ASE Vasp does not treat restart of a MD simulation well. Therefore, we run 
    directly from the command if INCAR aready exists.
    
    """
    import subprocess
    from ase.calculators.calculator import EnvironmentError, CalculationFailed

    try:
        proc = subprocess.Popen(command, shell=True, cwd=directory)
    except OSError as err:
        # Actually this may never happen with shell=True, since
        # probably the shell launches successfully.  But we soon want
        # to allow calling the subprocess directly, and then this
        # distinction (failed to launch vs failed to run) is useful.
        msg = 'Failed to execute "{}"'.format(command)
        raise EnvironmentError(msg) from err

    errorcode = proc.wait()

    if errorcode:
        path = os.path.abspath(directory)
        msg = ('Calculator "{}" failed with command "{}" failed in '
               '{} with error code {}'.format(name, command,
                                              path, errorcode))
        raise CalculationFailed(msg)

    return


@dataclasses.dataclass
class Cp2kStringReactorSetting(StringReactorSetting):

    backend: str = "cp2k"

    #: Number of tasks/processors/cpus for each image.
    ntasks_per_image: int = 1

    def __post_init__(self):
        """"""
        pairs = []

        pairs.extend(
            [
                ("GLOBAL", "RUN_TYPE BAND"),
                ("MOTION/BAND", "BAND_TYPE CI-NEB"),
                ("MOTION/BAND", f"NPROC_REP {self.ntasks_per_image}"),
                ("MOTION/BAND", f"NUMBER_OF_REPLICA {self.nimages}"),
                ("MOTION/BAND", f"K_SPRING {self.k/(units.Hartree/units.Bohr**2)}"),
                ("MOTION/BAND", "ROTATE_FRAMES F"),
                ("MOTION/BAND", "ALIGN_FRAMES F"),
                ("MOTION/BAND/CI_NEB", "NSTEPS_IT 2"),
                ("MOTION/BAND/OPTIMIZE_BAND", "OPT_TYPE DIIS"),
                ("MOTION/BAND/OPTIMIZE_BAND/DIIS", "NO_LS T"),
                ("MOTION/BAND/OPTIMIZE_BAND/DIIS", "N_DIIS 3"),
                ("MOTION/PRINT/RESTART_HISTORY/EACH", f"BAND {self.ckpt_period}"),
            ]
        )

        pairs.extend(
            [
                ("MOTION/PRINT/CELL", "_SECTION_PARAMETERS_ ON"),
                ("MOTION/PRINT/TRAJECTORY", "_SECTION_PARAMETERS_ ON"),
                ("MOTION/PRINT/FORCES", "_SECTION_PARAMETERS_ ON"),
            ]
        )
        self._internals["pairs"] = pairs

        return
    
    def get_run_params(self, *args, **kwargs):
        """"""
        # - convergence criteria
        rmax_ = kwargs.get("rmax", self.rmax)
        rrms_ = kwargs.get("rrms", self.rrms)
        fmax_ = kwargs.get("fmax", self.fmax)
        frms_ = kwargs.get("frms", self.frms)
        steps_ = kwargs.get("steps", self.steps)

        run_pairs = []
        run_pairs.append(
            ("MOTION/BAND/OPTIMIZE_BAND/DIIS", f"MAX_STEPS {steps_}")
        )
        if fmax_ is not None:
            run_pairs.extend(
                [
                    ("MOTION/BAND/CONVERGENCE_CONTROL", f"MAX_FORCE {fmax_/(units.Hartree/units.Bohr)}"),
                    ("MOTION/BAND/CONVERGENCE_CONTROL", f"MAX_DR {rmax_/(units.Bohr)}"),
                    ("MOTION/BAND/CONVERGENCE_CONTROL", f"RMS_FORCE {frms_/(units.Hartree/units.Bohr)}"),
                    ("MOTION/BAND/CONVERGENCE_CONTROL", f"RMS_DR {rrms_/(units.Bohr)}"),
                ]
            )

        # - add constraint
        run_params = dict(
            constraint = kwargs.get("constraint", self.constraint),
            run_pairs = run_pairs
        )

        return run_params


class Cp2kStringReactor(AbstractStringReactor):

    name: str = "cp2k"

    traj_name: str = "cp2k.out"

    def __init__(self, calc=None, params={}, ignore_convergence=False, directory="./", *args, **kwargs) -> None:
        """"""
        self.calc = calc
        if self.calc is not None:
            self.calc.reset()

        self.ignore_convergence = ignore_convergence

        self.directory = directory

        # - parse params
        self.setting = Cp2kStringReactorSetting(**params)
        self._debug(self.setting)

        return
    
    def _verify_checkpoint(self):
        """Check if the current directory has any valid outputs or 
            it just created the input files.

        """
        verified = super()._verify_checkpoint()
        if verified:
            checkpoints = list(self.directory.glob("*.restart"))
            self._debug(f"checkpoints: {checkpoints}")
            if not checkpoints:
                verified = False
        else:
            ...

        return verified
    
    def _irun(self, structures: List[Atoms], ckpt_wdir=None, *args, **kwargs):
        """"""
        try:
            if ckpt_wdir is None: # start from the scratch
                images = self._align_structures(structures)
                write(self.directory/"images.xyz", images)
                atoms = images[0] # use the initial state

                run_params = self.setting.get_run_params(**kwargs)
                run_params.update(**self.setting.get_init_params())

                # - update input template
                # GLOBAL section is automatically created...
                # FORCE_EVAL.(METHOD, POISSON)
                inp = self.calc.parameters.inp # string
                sec = parse_input(inp)
                for (k, v) in run_params["pairs"]:
                    sec.add_keyword(k, v)
                for (k, v) in run_params["run_pairs"]:
                    sec.add_keyword(k, v)

                # -- check constraint
                cons_text = run_params.pop("constraint", None)
                mobile_indices, frozen_indices = parse_constraint_info(
                    atoms, cons_text=cons_text, ignore_ase_constraints=True, ret_text=False
                )
                if frozen_indices:
                    #atoms._del_constraints()
                    #atoms.set_constraint(FixAtoms(indices=frozen_indices))
                    frozen_indices = sorted(frozen_indices)
                    sec.add_keyword(
                        "MOTION/CONSTRAINT/FIXED_ATOMS", 
                        "LIST {}".format(" ".join([str(i+1) for i in frozen_indices]))
                    )
            
                # -- add replica information
                band_section = sec.get_subsection("MOTION/BAND")
                for replica in images:
                    cur_rep = InputSection(name="REPLICA")
                    for pos in replica.positions:
                        cur_rep.add_keyword("COORD", ("{:.18e} "*3).format(*pos), unique=False)
                    band_section.subsections.append(cur_rep)
            else: # start from a checkpoint
                atoms  = read(ckpt_wdir/"images.xyz", "0")
                with open(ckpt_wdir/"cp2k.inp", "r") as fopen:
                    inp = "".join(fopen.readlines())
                sec = parse_input(inp)

                def remove_keyword(section, keywords):
                    """"""
                    for kw in keywords:
                        parts = kw.upper().split("/")
                        subsection = section.get_subsection("/".join(parts[0:-1]))
                        new_subkeywords = []
                        for subkw in subsection.keywords:
                            if parts[-1] not in subkw:
                                new_subkeywords.append(subkw)
                        subsection.keywords = new_subkeywords

                    return
                
                remove_keyword(
                    sec, 
                    keywords=[ # avoid conflicts
                        "GLOBAL/PROJECT", "GLOBAL/PRINT_LEVEL", "FORCE_EVAL/METHOD", 
                        "FORCE_EVAL/DFT/BASIS_SET_FILE_NAME", "FORCE_EVAL/DFT/POTENTIAL_FILE_NAME",
                        "FORCE_EVAL/SUBSYS/CELL/PERIODIC", 
                        "FORCE_EVAL/SUBSYS/CELL/A", "FORCE_EVAL/SUBSYS/CELL/B", "FORCE_EVAL/SUBSYS/CELL/C"
                    ]
                )

                sec.add_keyword("EXT_RESTART/RESTART_FILE_NAME", str(ckpt_wdir/"cp2k-1.restart"))
                sec.add_keyword("FORCE_EVAL/DFT/SCF/SCF_GUESS", "RESTART")

                # - copy wavefunctions...
                restart_wfns = sorted(list(ckpt_wdir.glob("*.wfn")))
                for wfn in restart_wfns:
                    (self.directory/wfn.name).symlink_to(wfn, target_is_directory=False)

            # - update input
            self.calc.parameters.inp = "\n".join(sec.write())
            atoms.calc = self.calc

            # - run calculation
            self.calc.atoms = atoms
            self.calc.write_input(atoms)
            run_cp2k("cp2k", self.calc.command, self.directory)
            self.calc.atoms = None

        except Exception as e:
            self._debug(e)
            self._debug(traceback.print_exc())

        return
    
    def read_convergence(self, *args, **kwargs):
        """"""
        converged = super().read_convergence(*args, **kwargs)

        with open(self.directory/"cp2k.out", "r") as fopen:
            lines = fopen.readlines()
        
        for line in lines:
            if "PROGRAM ENDED AT" in line:
                converged = True
                break

        return converged
    
    def _read_a_single_trajectory(self, wdir, *args, **kwargs):
        """

        NOTE: Fixed atoms have zero forces.

        """
        self._debug(f"***** read_trajectory *****")
        self._debug(f"{str(wdir)}")
        cell = None # TODO: if no pbc?
        natoms = None
        nimages = None
        temp_forces, temp_energies = [], []
        energies, forces = [], []
        with open(wdir/"cp2k.out", "r") as fopen:
            while True:
                line = fopen.readline()
                if not line:
                    break
                # - find cell
                if "CELL| Volume" in line:
                    found_cell = False
                    cell_data = []
                    for i in range(3):
                        line = fopen.readline()
                        if line:
                            cell_data.append(line)
                        else:
                            break
                    else:
                        found_cell = True
                    if found_cell:
                        try:
                            cell = [x.strip().split()[4:7] for x in cell_data]
                        except Exception as e:
                            self._debug("cell is not found.")
                            break
                # - find natoms
                if "TOTAL NUMBERS AND MAXIMUM NUMBERS" in line:
                    found_natoms = False
                    for i in range(3):
                        line = fopen.readline()
                        if not line:
                            break
                    else:
                        found_natoms = True
                    if found_natoms:
                        try:
                            natoms = int(line.strip().split()[-1])
                            self._debug(f"natoms: {natoms}")
                        except Exception as e:
                            self._debug("natoms is not found.")
                            break
                    else:
                        break
                if "Number of Images" in line:
                    # line = fopen.readline() # BUG: inconsistent Images and Replicas?
                    if not line:
                        break
                    try:
                        nimages = int(line.strip().split()[-2])
                        self._debug(line)
                        self._debug(f"nimages: {nimages}")
                    except Exception as e:
                        self._debug("nimages is not found.")
                # NOTE: For method with LineSearch, several SCF may be performed at one step
                """
                #if "Computing Energies and Forces" in line:
                if "REPLICA Nr." in line:
                    assert natoms is not None and nimages is not None, f"natoms: {natoms}, nimages: {nimages}"
                    curr_data = []
                    found_replica_forces = False
                    for i in range(natoms+2):
                        line = fopen.readline()
                        if line:
                            curr_data.append(line)
                        else:
                            break
                    else:
                        # current replica's forces are complete...
                        found_replica_forces = True
                    if found_replica_forces:
                        curr_energy = float(curr_data[0].strip().split()[-1])
                        energies.append(curr_energy)
                        curr_forces = [x.strip().split()[2:] for x in curr_data[2:]]
                        forces.append(curr_forces)
                    else:
                        break
                """
                if "Computing Energies and Forces" in line:
                    # NEB| REPLICA Nr.    1- Energy and Forces
                    # NEB|                                     Total energy:       -2940.286865478840
                    # NEB|    ATOM                            X                Y                Z
                    curr_data = []
                    found_replica_forces = False
                    for i in range((natoms+3)*nimages):
                        line = fopen.readline()
                        if line:
                            curr_data.append(line)
                        else:
                            break
                    else:
                        # current replica's forces are complete...
                        found_replica_forces = True
                    if found_replica_forces:
                        curr_energies = [
                            float(curr_data[i].strip().split()[-1]) for i in range(1, len(curr_data), natoms+3)
                        ]
                        temp_energies.append(curr_energies)
                        curr_forces = []
                        for ir in range(nimages):
                            curr_forces.append(
                                [curr_data[i].strip().split()[2:] for i in range((natoms+3)*ir+3, (natoms+3)*ir+3+natoms)]
                            )
                        temp_forces.append(curr_forces)
                    else:
                        break
                if "BAND TOTAL ENERGY" in line:
                    if temp_energies and temp_forces: # if the step completed...
                        #print("temp_energies: ", len(temp_energies))
                        #print("temp_forces: ", np.array(temp_forces, dtype=np.float64).shape)
                        energies.append(temp_energies[-1])
                        forces.extend(temp_forces[-1])
                        temp_forces, temp_energies = [], []

        # - truncate to the last complete band
        frames = [] # shape (nbands, nimages)
        if forces:
            forces = np.array(forces, dtype=np.float64)
            shape = forces.shape
            self._debug(f"forces: {shape}")
            nbands = int(shape[0]/nimages)
            forces = forces[:nbands*nimages]
            self._debug(f"truncated forces: {forces.shape} nbands: {nbands}")
            forces = np.reshape(forces, (nbands, nimages, natoms, -1)) # shape (nbands, nimages, natoms, 3)
            forces *= units.Hartree/units.Bohr

            energies = np.array(energies)[:nbands*nimages].reshape(nbands, nimages)
            energies *= units.Hartree
            self._debug(f"energies: {energies.shape} nbands: {nbands}")

            cell = np.array(cell, dtype=np.float64)
            self._debug(f"cell: {cell}")

            # - read positions
            frames_ = [] # shape (nimages, nbands)
            if nimages < 10:
                for i in range(nimages):
                    curr_xyzfile = wdir/f"cp2k-pos-Replica_nr_{i+1}-1.xyz"
                    curr_frames = read(curr_xyzfile, index=":", format="xyz")[:nbands]
                    frames_.append(curr_frames)
            else:
                for i in range(nimages):
                    curr_xyzfile = wdir/f"cp2k-pos-Replica_nr_{str(i+1).zfill(2)}-1.xyz"
                    curr_frames = read(curr_xyzfile, index=":", format="xyz")[:nbands]
                    frames_.append(curr_frames)
            for j in range(nbands):
                curr_band = []
                for i in range(nimages):
                    curr_band.append(frames_[i][j])
                frames.append(curr_band)
            for i in range(nbands):
                for j in range(nimages):
                    atoms = frames[i][j]
                    atoms.set_cell(cell)
                    atoms.pbc = True
                    spc = SinglePointCalculator(
                        atoms, energy=energies[i, j], free_energy=energies[i, j],
                        forces=forces[i, j].copy()
                    )
                    atoms.calc = spc
        else:
            ...

        return frames
    

if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List

from ase import Atoms
from ase.geometry import find_mic
from ase.calculators.singlepoint import SinglePointCalculator


def wrap_traj(frames: List[Atoms]):
    """Align positions according to the first frame.

    This is necessary for computing physical quantities base on atomic positions 
    with periodic boundary conditions.

    NOTE:
        This only works for fixed cell systems.
    
    TODO:
        Variable cell systems?

    """
    cell = frames[0].get_cell(complete=True)
    nframes = len(frames)
    for i in range(1,nframes):
        results = copy.deepcopy(frames[i].calc.results)
        prev_positions = frames[i-1].get_positions()
        curr_positions = frames[i].get_positions()
        shift = curr_positions - prev_positions
        curr_vectors, curr_distances = find_mic(shift, cell, pbc=True)
        frames[i].positions = prev_positions + curr_vectors
        frames[i].calc = SinglePointCalculator(frames[i], **results)

    return frames


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib

from ..placeholder import Placeholder
from ..variable import Variable
from .session import AbstractSession
from .utils import traverse_postorder


class Session(AbstractSession):

    def __init__(self, directory="./") -> None:
        """"""
        self.directory = pathlib.Path(directory)

        return

    def run(self, operation, feed_dict: dict = {}) -> None:
        """"""
        # - find forward order
        nodes_postorder = traverse_postorder(operation)
        for node in nodes_postorder:
            if hasattr(node, "_active") and node._active:
                node._active = False
                self._print(
                    f"Set {node} active to false as it is not supported in a basic session"
                )

        self._print(
            "[{:^24s}] NUM_NODES: {} AT MAIN: {}".format(
                "START", len(nodes_postorder), str(self.directory)
            )
        )

        # - run nodes
        for i, node in enumerate(nodes_postorder):
            # NOTE: reset directory since it maybe changed
            prev_name = node.directory.name
            if not prev_name:
                prev_name = node.__class__.__name__
            node.directory = self.directory / f"{str(i).zfill(4)}.{prev_name}"
            if node.__class__.__name__.endswith("Variable"):
                node_type = "VX"
            else:
                node_type = "OP"
            self._print(
                "[{:^24s}] NAME: {} AT {}".format(
                    node_type, node.__class__.__name__.upper(), node.directory.name
                )
            )

            if isinstance(node, Placeholder):
                node.output = feed_dict[node]
            elif isinstance(node, Variable):
                node.output = node.value
            else:  # Operation
                self._debug(f"node: {node}")
                self._process_operation(node)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
plot dimer curve 
"""

import argparse
from pathlib import Path

import numpy as np

import matplotlib
matplotlib.use('Agg') #silent mode
import matplotlib.pyplot as plt

from ase.io import read, write 

def parse_dimer_frames(xyzfile):
    frames = read(xyzfile, ':16') 
    dimer_symbols = frames[0].get_chemical_symbols()

    data = []
    for atoms in frames:
        assert len(atoms) == 2 
        energy = atoms.get_potential_energy() 
        dist = np.linalg.norm(atoms[0].position-atoms[1].position)
        data.append([dist,energy])
    data = np.array(data) 

    return dimer_symbols, data 

def parse_dimer_from_files(prefix):
    distances, energies = [], []
    for p in Path.cwd().glob(prefix):
        vasprun = p / 'vasprun.xml'
        atoms = read(vasprun, format='vasp-xml')
        distances.append(np.linalg.norm(atoms[0].position-atoms[1].position))
        energies.append(atoms.get_potential_energy())

    return distances, energies

def harvest_dimer_from_files(prefix):
    dimer_dirs = []
    for p in Path.cwd().glob(prefix):
        dimer_dirs.append(str(p))

    dimer_dirs.sort()

    dimer_frames = []
    for p in dimer_dirs:
        p = Path(p)
        vasprun = p / 'vasprun.xml'
        atoms = read(vasprun, format='vasp-xml')
        dimer_frames.append(atoms)

    write('dimer.xyz', dimer_frames)

    return 

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-p', '--prefix', 
        help='time series files'
    )

    args = parser.parse_args()

    harvest_dimer_from_files('Pt2-*')
    exit()

    #symbols, data = parse_dimer_frames('./evaluated.xyz')
    #distances, energies = parse_dimer_from_files('O2-*')
    distances, energies = parse_dimer_from_files(args.prefix)

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))
    ax.set_title(
        'Dimer', 
        fontsize=20, 
        fontweight='bold'
    )
    
    ax.set_xlabel('Distance [Å]', fontsize=16)
    ax.set_ylabel('Energyr [eV]', fontsize=16)

    ax.scatter(distances, energies)

    plt.savefig('dimer.png')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import numpy as np

from ase.io import read, write
from ase.constraints import FixAtoms, FixBondLengths
from ase.calculators.emt import EMT
from ase.neb import NEB
#from ase.neb import SingleCalculatorNEB
from ase.optimize import BFGS
from ase.geometry import wrap_positions
from sklearn.preprocessing import StandardScaler
from torch import conv_transpose1d


class AccCons():

    """ initialsie random placements
        Constrained Structure Swarm Exploration
    """

    """ fort.188
        1
        3
        6
        4
        0.04
        1    3    1.766034
        0
    """

    default_parameters = dict(
        rep_infor = 3,
        precon = 4,
        step_length = 0.04,
        init_delay_times = 20,
        basic_delay = 5,
        delay_plus = 4,
        neighbourhood_radius = 1.6,
        perturbation_factor = 0.09579038,
        max_ndishis = 10
    )

    fmax = 0.05

    # indicators
    imark = 0
    imark_att = 0
    ndishis = 0 # current number of distance history
    fix_times = 0
    delaycount = 0

    # data
    old_dis = None
    cur_dis = None

    old_energy = None
    cur_energy = None

    old_forces = None
    cur_force = None

    # 0 - balance
    old_bond_status = 0
    old_force_status = 0

    cur_bond_status = 0
    cur_force_status = 0

    def __init__(
        self, atoms, cpairs: list,
        **kwargs
    ):

        self.atoms = atoms
        self.cpairs = cpairs

        # init dynamics
        self.dyn = BFGS(self.atoms, maxstep=0.2) # maxstep = POTIM

        # init constrain
        for key, value in self.default_parameters.items():
            new_value = kwargs.get(key, None)
            if new_value is not None:
                value = new_value
            setattr(self, key, value)

        return
    
    def __call__(self, maxiters=20):
        """"""
        # start optimization
        # niter = 0
        self.stopsearch = False
        self.delaycount = self.init_delay_times
        self.updated_records = []
        self.constrain_records = []

        # opt params
        isconstrain = True
        #isconstrain = False

        #iswrap = True
        iswrap = False

        # init system info
        self.init_constrain_info() # not calculation

        # wrap atoms in the first place
        scaled_positions = self.atoms.get_scaled_positions()
        pos_a, pos_b = scaled_positions[self.cpairs[0]].copy(), scaled_positions[self.cpairs[1]].copy()
        conscentre = 0.5*(pos_a + pos_b)
        poses_wrap = wrap_positions(self.atoms.get_positions().copy(), self.atoms.cell, pbc=True, center=conscentre)
        self.atoms.set_positions(poses_wrap)

        write("188.xyz", self.atoms)

        for istep in range(maxiters):
            print(f"\n\n===== step {istep} =====")
            # calculation
            # log current info and single point
            self.update_maxforces_information("----- Force Information -----\n")
            # bfgs optimization
            self.dyn.step(self.cur_forces)
            poses_opt = self.atoms.get_positions()
            print("first 0: ", poses_opt[0])
            # move atoms in cube 000
            # wrap_positions
            if iswrap:
                # maybe wrong with Hessian and dr
                scaled_positions = self.atoms.get_scaled_positions()
                print("pos 0: ", self.atoms.get_positions()[0])
                print("pos 0: ", np.dot(scaled_positions, self.atoms.cell.complete())[0])
                print(scaled_positions[0])
                pos_a, pos_b = scaled_positions[self.cpairs[0]].copy(), scaled_positions[self.cpairs[1]].copy()
                conscentre = 0.5*(pos_a + pos_b)
                print("cons: ", conscentre)
                poses_opt = wrap_positions(poses_opt.copy(), self.atoms.cell, pbc=True, center=conscentre)
                print("pos_opt: ", poses_opt[0])
                #poses_opt = self.atoms.get_positions(wrap=True, pbc=True, center=conscentre).copy()
            if isconstrain:
                # transfer info before opt to old info
                self.update_his_constrain_info()
                # because of opt
                self.cur_poses = poses_opt
                self.cur_dis = self.atoms.get_distance(*self.cpairs, mic=True, vector=False)
                print("cur_dis: {} fix_dis: {} opt_dis: {}".format(self.old_dis, self.fixed_dis, self.cur_dis))
                # fix distance and get fixed_distance
                # ----- ----- ----- ----- -----
                self.check_constrain_conditions("----- Check Constrain Conditions -----\n")
                # constrain positions
                poses_fix = self.constrain_distance(
                    poses_opt.copy(), self.cur_forces.copy()
                )
                poses_opt = poses_fix
                # if update hessian
                # set new positions 
                # ----- ----- ----- ----- -----
            self.atoms.set_positions(poses_opt) # ready to call QM calculation

            write("188.xyz", self.atoms, append=True)

        return

    def init_constrain_info(self):
        """ init constrain info
        """
        self.old_poses = self.atoms.get_positions()
        self.old_dis = self.atoms.get_distance(*self.cpairs, mic=True, vector=False)

        self.cur_poses = self.old_poses.copy()
        self.cur_dis = self.old_dis

        natoms = len(self.atoms)
        self.old_forces = np.zeros((natoms,3))
        self.old_energy = 0.0
        self.cur_forces = np.zeros((natoms,3))
        self.cur_energy = 0.0

        self.fixed_dis = self.old_dis.copy()

        # important
        self.predicted_array = np.zeros((2,self.max_ndishis))

        content = '='*100 + '\n'
        content += 'Input System Information,\n'
        #content += 'System of %d atoms, 2 constrained, %d free, %d frozen.\n'\
        #        %(natoms, self.catoms.nfreeatoms, \
        #        self.catoms.nfrozenatoms)
        content += "A index <- {}, B index <- {}.\n".format(*self.cpairs)
        content += 'Distance between A and B is %.4f Ang.\n' %self.cur_dis
        content += '{:<50}{:<50}\n'.format('POSITIONS', 'FORCES')
        #for pos, force in zip(self.cur_poses, self.cur_forces):
        #    pos_line, force_line = '  ', '  '
        #    for j in range(3):
        #        pos_line += '{:>16}'.format(round(pos[j],8))
        #        force_line += '{:>16}'.format(round(force[j],8))
        #    content += pos_line + force_line + '\n'
        content += '='*100 + '\n'
        print(content)

        return

    def get_bond_status(self, d1, d2):
        """"""
        if d2 < d1:
            bond_status = -1 # attractive
        else:
            bond_status = 1 # repulsive

        if np.abs(d2 - d1) <= 1e-20:
            bond_status = 0 # balanced

        return bond_status
    
    # Methods for bond status
    def get_constrained_bond_status(self, poses, forces):
        """
        Description:
            +1 - attractive, -1 - repulsive, 0 - balanced.
        """
        #
        eps = np.ones(3)*self.perturbation_factor

        index_a, index_b = self.cpairs

        # get positions and forces of a and b
        pos_a, for_a = poses[index_a], forces[index_a]
        pos_b, for_b = poses[index_b], forces[index_b]

        #
        d1 = np.linalg.norm(pos_a - pos_b)
        
        # perturbation
        pos_a = pos_a + for_a*eps
        pos_b = pos_b + for_b*eps

        #
        d2 = np.linalg.norm(pos_a - pos_b) # TODO: minimum image

        return self.get_bond_status(d1, d2)
    
    def update_maxforces_information(self, marksentence):
        """
        Description:
            Update the largest and second largest forces among all atoms, 
            along with their indices. The largest partial force among A 
            and B is also given.
        Called By: 
            self.cbfgs()
        IN:
            index_a, index_b, natoms, forces.
        OUT:
            self.cformax,
            self.formax1_index, self.formax2_index, 
            self.formax1_vec, self.formax2_vec, 
            self.formax1_norm, self.formax2_vec.
        """

        # ----- update current constrain info -----
        # from QM calculation
        self.cur_poses = self.atoms.get_positions()
        self.cur_forces = self.atoms.get_forces().copy() # forcecall
        self.cur_energy = self.atoms.get_potential_energy()

        self.cur_dis = self.atoms.get_distance(*self.cpairs, mic=True, vector=False)

        self.cur_bond_status = self.get_bond_status(self.old_dis, self.cur_dis)
        self.cur_force_status = self.get_constrained_bond_status(
            self.cur_poses, self.cur_forces
        )
        
        # ----- update forces info -----
        index_a, index_b = self.cpairs

        natoms = len(self.atoms)
        forces = self.cur_forces

        # find force info
        f1_index, f2_index = 0, 0
        f1_vec, f2_vec = np.array([]), np.array([])
        f1_norm, f2_norm = 0.0, 0.0

        for i in range(natoms):
            f_vec = forces[i]
            f_norm = np.linalg.norm(max(np.fabs(f_vec)))
            if f_norm > f1_norm:
                # move 1st to 2nd
                f2_norm, f2_index, f2_vec = f1_norm, f1_index, f1_vec
                # update 1st
                f1_norm, f1_index, f1_vec = f_norm, i, f_vec
            elif f_norm > f2_norm:
                f2_norm, f2_index, f2_vec = f_norm, i, f_vec

        # update
        self.cformax = max(max(forces[index_a]), max(forces[index_b]))
        self.formax1_index, self.formax2_index = f1_index, f2_index
        self.formax1_vec, self.formax2_vec = f1_vec, f2_vec
        self.formax1_norm, self.formax2_norm = f1_norm, f2_norm

        # log
        content = marksentence
        content += "Find the largest and second largest frac-force.\n"
        content += ("{:<15}  "*4+"\n").format('Max Force','Atom Index','Force Norm','Force Vector')
        content += ("{:<15}  "*2+"{:<15.4f}  "+"{:<.4f}  "*3+"\n").format(
            1,f1_index,np.linalg.norm(f1_vec),f1_vec[0],f1_vec[1],f1_vec[2]
        )
        content += ("{:<15}  "*2+"{:<15.4f}  "+"{:<.4f}  "*3+"\n").format(
            2,f2_index,np.linalg.norm(f2_vec),f2_vec[0],f2_vec[1],f2_vec[2]
        )
        print(content)

        return
    
    def update_his_constrain_info(self):
        """"""
        self.old_poses = self.cur_poses.copy()
        self.old_forces = self.cur_forces.copy()

        self.old_energy = self.cur_energy

        self.old_dis = self.cur_dis

        self.old_bond_status = self.cur_bond_status
        self.old_force_status = self.cur_force_status

        return

    
    def constrain_distance(self, poses_in, forces):
        """
        Description:
            Get atoms' positions after constrain.
        """
        poses = poses_in.copy()
        # get work variables
        natoms = poses.shape[0]
        index_a, index_b = self.cpairs

        # positions and forces
        cur_dis, fixed_dis = self.cur_dis, self.fixed_dis

        # get positions and forces of a and b
        pos_a, for_a = poses[index_a].copy(), forces[index_a].copy()
        pos_b, for_b = poses[index_b].copy(), forces[index_b].copy()

        # wrong !!! array object change in memory
        #pos_a, for_a = poses[index_a], forces[index_a].copy()
        #pos_b, for_b = poses[index_b], forces[index_b].copy()

        # calculate new positions of a and b
        lamb = -(np.linalg.norm(for_a) - np.linalg.norm(for_b)) / (np.linalg.norm(for_a) + np.linalg.norm(for_b))
        pos_a_new = (pos_a + pos_b) / 2.0 + \
                (lamb + fixed_dis/cur_dis - lamb*fixed_dis/cur_dis) * \
                (pos_a - pos_b) / 2.0
        pos_b_new = (pos_a + pos_b) / 2.0 + \
                (lamb - fixed_dis/cur_dis - lamb*fixed_dis/cur_dis) * \
                (pos_a - pos_b) / 2.0

        # move neighbours and log
        """
        radius = self.neighbourhood_radius
        poses, content = self.catoms.move_nearest_neighbours(
            natoms, poses,
            pos_a, pos_a_new, index_a, radius
        ) # move a's neighbours
        print(content)

        poses, content = self.catoms.move_nearest_neighbours(
            natoms, poses, \
            pos_b, pos_b_new, index_b, radius
        ) # move a's neighbours
        print(content)
        """

        # set new positions
        poses[index_a] = pos_a_new
        poses[index_b] = pos_b_new

        return poses

    def check_constrain_conditions(self, marksentence):
        """ constrain conditions
        """
        # force condition
        index_a, index_b = self.cpairs
        cond1 = (
            self.formax1_index == index_a and 
            self.formax2_index == index_b
        ) or (
            self.formax1_index == index_b and
            self.formax2_index == index_a
        )

        # delay condition
        cond2 = (self.delaycount <= 0)

        # consistent condition
        cond3 = (self.cur_bond_status * self.cur_force_status >= 0)

        # convergence condition
        cond4 = self.cformax > self.fmax

        # log
        content = marksentence
        content += '%6s-if constrained atoms have largest forces \n' %cond1
        content += '%6s-if delaycount is smaller than zero \n' %cond2
        content += '%6s-if bond status is consistent with force status \n' %cond3
        content += '%6s-if forces on constrained atoms are larger than ftol\n ' \
                %cond4
        print(content)

        # get predicted distance
        self.predicted_dis = self.get_predicted_distance()

        # change parameters
        old_dis, cur_dis = self.old_dis, self.cur_dis

        if not cond4:
            self.precon = 0.05

        isfixed = False
        if self.cur_bond_status == 1:
            # log
            print("NOTICE: Bond is repulsive, try to decrease distance.\n")

            # check conditions
            if cond1 and cond2 and cond3 and cond4:
                print('Now delay is done, record information.\n')
                # auto search
                self.rep_infor = self.rep_infor + 1
                print('Current RepTag is %d. Once 6, decrease distance.\n' \
                        %self.rep_infor)
                print('*'*100+'\n')

                # reach rep times
                if self.rep_infor == 6:
                    # initialize self.rep_infor
                    self.rep_infor = 3
                    # very close to TS
                    self.check_predicted_distance('repulsive')
                    # update
                    self.updated_records.append({'fix_times': self.fix_times, \
                            'distance_in': self.old_dis, \
                            'distance_out': self.predicted_dis})
                    self.fix_times = 0
                else:
                    self.fix_times += 1
            else:
                isfixed = True

        elif self.cur_bond_status == -1:
            # log
            print('NOTICE: Bond is attractive, try to increase distance.\n')

            if cond1 and cond2 and cond3 and cond4:
                # log
                self.rep_infor -= 1
                print('Current RepTag is %d. Once 0, increase distance.\n' \
                        %self.rep_infor)
                print('*'*100+'\n')
                if self.rep_infor == 0:
                    self.rep_infor = 3
                    # very close to TS
                    self.check_predicted_distance('attractive')
                    # update
                    self.updated_records.append({'fix_times': self.fix_times, \
                            'distance_in': self.old_dis, \
                            'distance_out': self.predicted_dis})
                    self.fix_times = 0
                else:
                    self.fix_times += 1
            else:
                isfixed = True

        elif self.cur_bond_status == 0:
            print('NOTICE: Bond is banlanced, do nothing.\n')
            isfixed = True
        
        if isfixed:
            # update
            self.fix_times += 1
            
            # log
            content = '*'*100+'\n' 
            content += 'Constrain: Remain distance.\n' 
            content += 'Fixed Distance: %.4f at %d delay times.\n' \
                    %(self.fixed_dis, self.delaycount)
            content += '*'*100+'\n'
            print(content)

        # set delaycount
        self.delaycount -= 1
    
        # distance
    def get_predicted_distance(self):
        """
        use history to predict distance
        Here, old_dis - input, cur_dis - after opt
        """
        #
        old_dis, cur_dis = self.old_dis, self.cur_dis
        step_length = self.step_length

        # array 0 1 2 3 4 5 6 7 8 9
        # dis   new <- old
        predicted_array = self.predicted_array

        # update predicted array
        self.ndishis += 1
        if self.ndishis >= self.max_ndishis:
            self.ndishis = self.max_ndishis
        if self.ndishis > 1:
            predicted_array = self.shift_predicted_array(predicted_array)

        p1 = self.formax1_norm**2
        p2 = old_dis - (cur_dis - old_dis)*(1.0 + step_length*10.0)
        predicted_array[0,0] = p1 
        predicted_array[1,0] = p2

        #  get predicted distance
        #predicted_dis = old_dis
        for nhis in range(self.ndishis,0,-1):
            dtmp, dtmp1 = 0.0, 0.0
            for i in range(nhis):
                dtmp += 1.0/predicted_array[0,i]
            for i in range(nhis):
                dtmp1 += 1.0/predicted_array[0,i]/dtmp*predicted_array[1,i]
            # check consistent
            if (old_dis - dtmp1)*(cur_dis - old_dis) > 0:
                if np.abs(dtmp1 - old_dis) < step_length:
                    predicted_dis = dtmp1
                else:
                    predicted_dis = old_dis + (dtmp1 - old_dis) / \
                            np.abs(dtmp1 - old_dis) * step_length
                break
        else:
            predicted_dis = old_dis - (cur_dis - old_dis) * \
                    (1.0 + step_length*10.0)
        
        # refine predicted distance
        ddis = predicted_dis - old_dis
        if ddis >= 0:
            ddis_sign = 1.0
        else:
            ddis_sign = -1.0

        if abs(ddis) < 0.01 and predicted_array[0,0] > 0.01:
            predicted_dis = old_dis + ddis_sign * \
                    min(max(abs(ddis*10), 0.005), 0.02) * step_length/0.05
            print('Small distance change.\n')

        ddis = predicted_dis - old_dis
        if abs(ddis) < 0.0025:
            predicted_dis = old_dis + ddis/abs(ddis)*0.0025

        # log
        content = 'Current constrained distance is %.4f.\n' %self.fixed_dis
        content += 'Has already collected %d history.\n' %self.ndishis
        content += 'Use %d distance history. Predicted Distance: %.4f Ang\n' \
                %(nhis, predicted_dis)
        print(content)

        # update
        self.predicted_array = predicted_array
        self.ndishis = nhis

        return predicted_dis
        # return self.old_dis**2 / self.cur_dis
    
    def shift_predicted_array(self, predicted_array):
        """"""
        max_ndishis = self.max_ndishis
        for i in range(1, max_ndishis):
            predicted_array[0,i] = predicted_array[0,i-1]
            predicted_array[1,i] = predicted_array[1,i-1]
        return predicted_array
    
    def check_predicted_distance(self, status_tag):
        """change search parameters to give more subtle performance."""
        # get marks
        if status_tag == 'repulsive':
            mark1, mark2 = self.imark, self.imark_att 
        elif status_tag == 'attractive':
            mark1, mark2 = self.imark_att, self.imark
        else:
            raise ValueError('Wrong status tag.')

        # update distance
        dtmp = self.fixed_dis # distance before opt
        # overwrite codes for update distance
        self.fixed_dis = self.predicted_dis
        if abs(dtmp - self.fixed_dis) > 0.005:
            self.ireset_hessian = 1

        # log
        content = 'Constrain: Update distance.\n'
        content += 'Successfully change distance from %.4f to %.4f.\n' \
                %(dtmp, self.fixed_dis)
        content += 'Predicted Distance: %.4f at %d fixed times \
        and %d delay times.\n' \
                %(self.predicted_dis, self.fix_times, self.delaycount)

        # update
        mark1 += 1 # change tag

        # change precon
        if self.precon != 2:
            self.delaycount = self.basic_delay + self.delay_plus
        self.precon /= 2.0
        if self.precon < 0.005:
            self.precon = 0.01

        # refine precon, which controls the step length for bond change
        # check if bond is oscillating between attractive and repulsive
        if mark2 >= 1:
            content += 'Overshoot! Too %s the bond.\n' %status_tag
            mark2 = 0
        # reduce precon after continuing increase or decrease
        if mark1 == 2 and self.precon == 0.5:
            content += 'Remain %s after change at precon==2.\n' %status_tag
            mark1 = 0
            self.precon = 2
        # still after 3 changes, give more delay times to find MEP
        if mark1 == 3:
            content += 'still %s after 3 changes.\n' %status_tag
            mark1 = 0
            self.ireset_hessian = 1
            self.delaycount = self.basic_delay + self.delay_plus
            if self.cformax > 0.1:
                self.precon = 0.4
            else:
                self.delaycount = self.basic_delay + self.delay_plus + 1

        content += 'Delay times %d. Precon %.4f\n' \
                %(self.delaycount, self.precon)

        # log
        content += '*'*100+'\n'
        print(content)

        # update marks
        if status_tag == 'repulsive':
            self.imark, self.imark_att = mark1, mark2
        elif status_tag == 'attractive':
            self.imark_att, self.imark = mark1, mark2
        else:
            raise ValueError('Wrong status tag.')

        return


def test_cons(atoms, calc):
    atoms.calc = calc
    atoms.set_constraint(
        FixBondLengths(
            pairs = [[0, 23]],
            bondlengths = [2.426]
        )
    )

    qn = BFGS(atoms, trajectory="cons.traj")
    qn.run(fmax=0.05, steps=50)

    return



if __name__ == "__main__":
    # test LASP-NEB
    from gdpx.computation.lasp import LaspNN
    pot_path = "/mnt/scratch2/users/40247882/catsign/lasp-main/ZnCrOCH.pot"
    pot = dict(
        C  = pot_path,
        O  = pot_path,
        Cr = pot_path,
        Zn = pot_path
    )
    calc = LaspNN(
        directory = "./LaspNN-Worker",
        command = "mpirun -n 4 lasp",
        pot=pot
    )

    initial = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-IS.xyz")
    transition = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-TS.xyz")
    final = read("/mnt/scratch2/users/40247882/catsign/ZnOx/P188/CO-FS.xyz")

    transition = read("/mnt/scratch2/users/40247882/catsign/ZnOx/cases/TStest.xyz")

    cons_indices = list(range(1,13)) + list(range(24,36))
    print("indices: ", cons_indices)
    constraint = FixAtoms(indices=cons_indices) # first atom is O
    #constraint = FixAtoms(indices=[1,2,3,4,5,6,7,8]) # first atom is O

    transition.set_constraint(constraint)
    transition.calc = calc

    #dyn = BFGS(transition, maxstep=0.2) # maxstep = POTIM
    #dyn.run(0.05, 200)
    #exit()

    maxiters = int(sys.argv[1])

    # test_cons(transition, calc)
    constrdyn = AccCons(transition, cpairs=[0,23])
    constrdyn(maxiters)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from typing import Mapping

from ..data.array import AtomsNDArray
from .validator import AbstractValidator

class RankValidator(AbstractValidator):

    def run(self, dataset: Mapping[str, AtomsNDArray], *args, **kwargs):
        """"""
        super().run(*args, **kwargs)

        # -
        prediction = dataset["prediction"].get_marked_structures()
        reference = dataset["reference"].get_marked_structures()
        self._print(prediction)
        self._print(reference)

        return


if __name__ == "__main__":
    ...